"""
Ideally should be in the `bases` module, however we'll run into a Circular
Import scenario if we move it there, since the `loaders` and `dumpers` modules
both import directly from `bases`.

"""
from __future__ import annotations

import logging
import warnings
from datetime import datetime, date
from typing import Mapping

from .bases import AbstractMeta, META, AbstractEnvMeta
from .class_helper import (
    META_INITIALIZER, _META, get_meta,
    get_outer_class_name, get_class_name, create_new_class,
    json_field_to_dataclass_field, dataclass_field_to_json_field,
    field_to_env_var,
)
from .decorators import try_with_load
from .enums import DateTimeTo, LetterCase, LetterCasePriority
from .errors import ParseError, show_deprecation_warning
from .loader_selection import get_dumper, get_loader
from .log import LOG
from .type_def import E
from .utils.type_conv import date_to_timestamp, as_enum


# global flag to determine if debug mode was ever enabled
_debug_was_enabled = False


def register_type(cls, tp, *, load=None, dump=None) -> None:
    from .dumpers import DumpMixin
    from .loaders import LoadMixin

    dumper = get_dumper(cls, base_cls=DumpMixin)
    loader = get_loader(cls, base_cls=LoadMixin)

    # default hooks
    load = tp if load is None else load
    dump = str if dump is None else dump

    # adapt to what v0 expects
    load = _adapt_to_arity(load, loader.HOOK_ARITY)
    dump = _adapt_to_arity(dump, dumper.HOOK_ARITY)

    dumper.register_dump_hook(tp, dump)
    loader.register_load_hook(tp, load)


# use `debug_enabled` for log level if it's a str or int.
def _enable_debug_mode_if_needed(cls_loader, possible_lvl):
    global _debug_was_enabled
    if not _debug_was_enabled:
        _debug_was_enabled = True
        # use `debug_enabled` for log level if it's a str or int.
        default_lvl = logging.DEBUG
        # minimum logging level for logs by this library.
        min_level = default_lvl if isinstance(possible_lvl, bool) else possible_lvl
        # set the logging level of this library's logger.
        LOG.setLevel(min_level)
        LOG.info('DEBUG Mode is enabled')

    # Decorate all hooks so they format more helpful messages
    # on error.
    load_hooks = cls_loader.__LOAD_HOOKS__
    for typ in load_hooks:
        load_hooks[typ] = try_with_load(load_hooks[typ])


def _as_enum_safe(cls: type, name: str, base_type: type[E]) -> 'E | None':
    """
    Attempt to return the value for class attribute :attr:`attr_name` as
    a :type:`base_type`.

    :raises ParseError: If we are unable to convert the value of the class
      attribute to an Enum of type `base_type`.
    """
    try:
        return as_enum(getattr(cls, name), base_type)

    except ParseError as e:
        # We run into a parsing error while loading the enum; Add
        # additional info on the Exception object before re-raising it
        e.class_name = get_class_name(cls)
        e.field_name = name
        raise


def _arity(hook) -> int:
    # Python function / method
    code = getattr(hook, "__code__", None)
    if code is not None:
        # reject *args/**kwargs if you want strictness
        if code.co_flags & 0x04 or code.co_flags & 0x08:
            return -1
        return code.co_argcount

    # Classes / C-callables (e.g., IPv4Address) don't expose __code__.
    # Treat as "callable(value)" i.e., 1-arg constructor.
    return 1


def _adapt_to_arity(fn, target_arity: int):
    src = _arity(fn)

    if src == -1:
        # If they already accept *args/**kwargs, it will work everywhere.
        return fn

    if src == target_arity:
        return fn

    # Common case: user gives 1-arg callable but backend passes extra info
    if src == 1 and target_arity > 1:
        def wrapper(x, *rest):
            return fn(x)
        return wrapper

    # Less common: user gives 2-arg (v1 codegen) but v0 expects 1
    # You can reject this unless you have a sane mapping.
    raise TypeError(
        f"Hook {getattr(fn, '__name__', fn)!r} has {src} args, "
        f"but backend expects {target_arity}."
    )


class BaseJSONWizardMeta(AbstractMeta):
    """
    Superclass definition for the `JSONWizard.Meta` inner class.

    See the implementation of the :class:`AbstractMeta` class for the
    available config that can be set, as well as for descriptions on any
    implemented methods.
    """

    __slots__ = ()

    @classmethod
    def _init_subclass(cls):
        """
        Hook that should ideally be run whenever the `Meta` class is
        sub-classed.

        """
        outer_cls_name = get_outer_class_name(cls, raise_=False)

        # We can retrieve the outer class name using `__qualname__`, but it's
        # not easy to find the class definition itself. The simplest way seems
        # to be to create a new callable (essentially a class method for the
        # outer class) which will later be called by the base enclosing class.
        #
        # Note that this relies on the observation that the
        # `__init_subclass__` method of any inner classes are run before the
        # one for the outer class.
        if outer_cls_name is not None:
            META_INITIALIZER[outer_cls_name] = cls.bind_to
        else:
            from .abstractions import AbstractJSONWizard

            # The `Meta` class is defined as an outer class. Emit a warning
            # here, just so we can ensure awareness of this special case.
            LOG.warning('The %r class is not declared as an Inner Class, so '
                        'these are global settings that will apply to all '
                        'JSONSerializable sub-classes.', get_class_name(cls))

            # Copy over global defaults to the :class:`AbstractMeta`
            for attr in AbstractMeta.fields_to_merge:
                setattr(AbstractMeta, attr, getattr(cls, attr, None))
            if cls.json_key_to_field:
                AbstractMeta.json_key_to_field = cls.json_key_to_field

            # Create a new class of `Type[W]`, and then pass `create=False` so
            # that we don't create new loader / dumper for the class.
            new_cls = create_new_class(cls, (AbstractJSONWizard, ))
            cls.bind_to(new_cls, create=False)

    @classmethod
    def bind_to(cls, dataclass: type, create=True, is_default=True,
                base_loader=None,
                base_dumper=None):
        cls_loader = get_loader(dataclass, create=create,
                                base_cls=base_loader)
        cls_dumper = get_dumper(dataclass, create=create,
                                base_cls=base_dumper)

        if cls.debug_enabled:
            _enable_debug_mode_if_needed(cls_loader, cls.debug_enabled)

        if cls.json_key_to_field is not None:
            add_for_both = cls.json_key_to_field.pop('__all__', None)

            json_field_to_dataclass_field(dataclass).update(
                cls.json_key_to_field
            )

            if add_for_both:
                dataclass_to_json_field = dataclass_field_to_json_field(
                    dataclass)

                # We unfortunately can't use a dict comprehension approach, as
                # we don't know if there are multiple JSON keys mapped to a
                # single dataclass field. So to be safe, we should only set
                # the first JSON key mapped to each dataclass field.
                for json_key, field in cls.json_key_to_field.items():
                    if field not in dataclass_to_json_field:
                        dataclass_to_json_field[field] = json_key

        if cls.marshal_date_time_as is not None:
            enum_val = _as_enum_safe(cls, 'marshal_date_time_as', DateTimeTo)

            if enum_val is DateTimeTo.TIMESTAMP:
                # Update dump hooks for the `datetime` and `date` types
                cls_dumper.dump_with_datetime = lambda o, *_: round(o.timestamp())
                cls_dumper.dump_with_date = lambda o, *_: date_to_timestamp(o)
                cls_dumper.register_dump_hook(
                    datetime, cls_dumper.dump_with_datetime)
                cls_dumper.register_dump_hook(
                    date, cls_dumper.dump_with_date)

            elif enum_val is DateTimeTo.ISO_FORMAT:
                # noop; the default dump hook for `datetime` and `date`
                # already serializes using this approach.
                pass

        if cls.key_transform_with_load is not None:
            cls_loader.transform_json_field = _as_enum_safe(
                cls, 'key_transform_with_load', LetterCase)

        if cls.key_transform_with_dump is not None:
            cls_dumper.transform_dataclass_field = _as_enum_safe(
                cls, 'key_transform_with_dump', LetterCase)

        # Finally, if needed, save the meta config for the outer class. This
        # will allow us to access this config as part of the JSON load/dump
        # process if needed.
        if is_default:
            # Check if the dataclass already has a Meta config; if so, we need to
            # copy over special attributes so they don't get overwritten.
            if dataclass in _META:
                _META[dataclass] &= cls
            else:
                _META[dataclass] = cls


class BaseEnvWizardMeta(AbstractEnvMeta):
    """
    Superclass definition for the `EnvWizard.Meta` inner class.

    See the implementation of the :class:`AbstractEnvMeta` class for the
    available config that can be set, as well as for descriptions on any
    implemented methods.
    """

    __slots__ = ()

    @classmethod
    def _init_subclass(cls):
        """
        Hook that should ideally be run whenever the `Meta` class is
        sub-classed.

        """
        outer_cls_name = get_outer_class_name(cls, raise_=False)

        if outer_cls_name is not None:
            META_INITIALIZER[outer_cls_name] = cls.bind_to
        else:
            from .abstractions import AbstractJSONWizard

            # The `Meta` class is defined as an outer class. Emit a warning
            # here, just so we can ensure awareness of this special case.
            LOG.warning('The %r class is not declared as an Inner Class, so '
                        'these are global settings that will apply to all '
                        'EnvWizard sub-classes.', get_class_name(cls))

            # Copy over global defaults to the :class:`AbstractMeta`
            for attr in AbstractEnvMeta.fields_to_merge:
                setattr(AbstractEnvMeta, attr, getattr(cls, attr, None))
            if cls.field_to_env_var:
                AbstractEnvMeta.field_to_env_var = cls.field_to_env_var

            # Create a new class of `Type[W]`, and then pass `create=False` so
            # that we don't create new loader / dumper for the class.
            new_cls = create_new_class(cls, (AbstractJSONWizard, ))
            cls.bind_to(new_cls, create=False)

    @classmethod
    def bind_to(cls, env_class: type, create=True, is_default=True):
        meta = get_meta(env_class)

        cls_loader = get_loader(
            env_class,
            create=create,
            env=True)
        cls_dumper = get_dumper(
            env_class,
            create=create)

        if cls.debug_enabled:
            _enable_debug_mode_if_needed(cls_loader, cls.debug_enabled)

        if cls.field_to_env_var is not None:
            field_to_env_var(env_class).update(
                cls.field_to_env_var
            )

        cls.key_lookup_with_load = _as_enum_safe(
            cls, 'key_lookup_with_load', LetterCasePriority)

        cls_dumper.transform_dataclass_field = _as_enum_safe(
            cls, 'key_transform_with_dump', LetterCase)

        # Finally, if needed, save the meta config for the outer class. This
        # will allow us to access this config as part of the JSON load/dump
        # process if needed.
        if is_default:
            # Check if the dataclass already has a Meta config; if so, we need to
            # copy over special attributes so they don't get overwritten.
            if env_class in _META:
                _META[env_class] &= cls
            else:
                _META[env_class] = cls


# noinspection PyPep8Naming
def LoadMeta(**kwargs) -> META:
    """
    Helper function to setup the ``Meta`` Config for the JSON load
    (de-serialization) process, which is intended for use alongside the
    ``fromdict`` helper function.

    For descriptions on what each of these params does, refer to the `Docs`_
    below, or check out the :class:`AbstractMeta` definition (I want to avoid
    duplicating the descriptions for params here).

    Examples::

        >>> LoadMeta(key_transform='CAMEL').bind_to(MyClass)
        >>> fromdict(MyClass, {"myStr": "value"})

    .. _Docs: https://dcw.ritviknag.com/en/latest/common_use_cases/meta.html
    """
    base_dict = kwargs | {'__slots__': ()}

    if (v := base_dict.pop('key_transform', None)) is not None:
        base_dict['key_transform_with_load'] = v

    # Create a new subclass of :class:`AbstractMeta`
    # noinspection PyTypeChecker
    return type('Meta', (BaseJSONWizardMeta, ), base_dict)


# noinspection PyPep8Naming
def DumpMeta(**kwargs) -> META:
    """
    Helper function to setup the ``Meta`` Config for the JSON dump
    (serialization) process, which is intended for use alongside the
    ``asdict`` helper function.

    For descriptions on what each of these params does, refer to the `Docs`_
    below, or check out the :class:`AbstractMeta` definition (I want to avoid
    duplicating the descriptions for params here).

    Examples::

        >>> DumpMeta(key_transform='CAMEL').bind_to(MyClass)
        >>> asdict(MyClass, {"myStr": "value"})

    .. _Docs: https://dcw.ritviknag.com/en/latest/common_use_cases/meta.html
    """

    # Set meta attributes here.
    base_dict = kwargs | {'__slots__': ()}

    if (v := base_dict.pop('key_transform', None)) is not None:
        base_dict['key_transform_with_dump'] = v

    # Create a new subclass of :class:`AbstractMeta`
    # noinspection PyTypeChecker
    return type('Meta', (BaseJSONWizardMeta, ), base_dict)


# noinspection PyPep8Naming
def EnvMeta(**kwargs) -> META:
    """
    Helper function to setup the ``Meta`` Config for the EnvWizard.

    For descriptions on what each of these params does, refer to the `Docs`_
    below, or check out the :class:`AbstractEnvMeta` definition (I want to avoid
    duplicating the descriptions for params here).

    Examples::

        >>> EnvMeta(key_transform_with_dump='SNAKE').bind_to(MyClass)

    .. _Docs: https://dcw.ritviknag.com/en/latest/common_use_cases/meta.html
    """

    # Set meta attributes here.
    base_dict = kwargs | {'__slots__': ()}

    # Create a new subclass of :class:`AbstractMeta`
    # noinspection PyTypeChecker
    return type('EnvMeta', (BaseEnvWizardMeta, ), base_dict)



import json

from .._dumpers import asdict
from .._loaders import fromdict, fromlist
from .._serial_json import JSONWizard
from ..utils.containers import Container


class JSONListWizard(JSONWizard):
    """
    A Mixin class that extends :class:`JSONWizard` to return
    :class:`Container` - instead of `list` - objects.

    Note that `Container` objects are simply convenience wrappers around a
    collection of dataclass instances. For all intents and purposes, they
    behave exactly the same as `list` objects, with some added helper methods:

        * ``prettify`` - Convert the list of instances to a *prettified* JSON
          string.

        * ``to_json`` - Convert the list of instances to a JSON string.

        * ``to_json_file`` - Serialize the list of instances and write it to a
          JSON file.

    """
    @classmethod
    def from_json(cls, string, *,
                  decoder=json.loads,
                  **decoder_kwargs):
        """
        Converts a JSON `string` to an instance of the dataclass, or a
        Container (list) of the dataclass instances.
        """
        o = decoder(string, **decoder_kwargs)

        if isinstance(o, dict):
            return fromdict(cls, o)

        return Container[cls](fromlist(cls, o))

    @classmethod
    def from_list(cls, o):
        """
        Converts a Python `list` object to a Container (list) of the dataclass
        instances.
        """
        return Container[cls](fromlist(cls, o))


class JSONFileWizard:
    """
    A Mixin class that makes it easier to interact with JSON files.

    This can be paired with the :class:`JSONWizard` Mixin
    class for more complete extensibility.

    """
    @classmethod
    def from_json_file(cls, file, *,
                       decoder=json.load,
                       **decoder_kwargs):
        """
        Reads in the JSON file contents and converts to an instance of the
        dataclass, or a list of the dataclass instances.
        """
        with open(file) as in_file:
            o = decoder(in_file, **decoder_kwargs)

        return fromdict(cls, o) if isinstance(o, dict) else fromlist(cls, o)

    def to_json_file(self, file, mode='w',
                     encoder=json.dump,
                     **encoder_kwargs):
        """
        Serializes the instance and writes it to a JSON file.
        """
        with open(file, mode) as out_file:
            encoder(asdict(self), out_file, **encoder_kwargs)



from typing import Any, Callable, TypeVar
from weakref import WeakKeyDictionary

from ._type_def import T

K = TypeVar('K')
V = TypeVar('V')

def per_cls(
    cache: WeakKeyDictionary[type, V],
    cls: type,
    factory: Callable[[], dict[K, V]] = dict,
) -> dict[K, V]: ...

def is_builtin(o: Any) -> bool:
    """Check if an object/singleton/class is a builtin in Python."""


def create_new_class(
        class_or_instance, bases: tuple[T, ...],
        suffix: str | None = None, attr_dict=None) -> T:
    """
    Create (dynamically) and return a new class that sub-classes from a list
    of `bases`.
    """


def get_class_name(class_or_instance) -> str:
    """Return the fully qualified name of a class."""


def get_outer_class_name(inner_cls, default=None, raise_: bool = True) -> str:
    """
    Attempt to return the fully qualified name of the outer (enclosing) class,
    given a reference to the inner class.

    If any errors occur - such as when `inner_cls` is not a real inner
    class - then an error will be raised if `raise_` is true, and if not
    we will return `default` instead.

    """


def get_class(obj: Any) -> type:
    """Get the class for an object `obj`"""


def is_subclass(obj: Any, base_cls: type) -> bool:
    """Check if `obj` is a sub-class of `base_cls`"""


def is_subclass_safe(cls, class_or_tuple) -> bool:
    """Check if `obj` is a sub-class of `base_cls` (safer version)"""



from dataclasses import dataclass

from dataclass_wizard import JSONWizard


# noinspection PyCompatibility
def test_union_as_type_alias_recursive():
    """
    Recursive or self-referential `Union` (defined as `TypeAlias`)
    types are supported.
    """
    type JSON = str | int | float | bool | dict[str, JSON] | list[JSON] | None

    @dataclass
    class MyTestClass(JSONWizard):
        name: str
        meta: str
        msg: JSON

    x = MyTestClass.from_dict(
        {
            "name": "name",
            "meta": "meta",
            "msg": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}],
        }
    )
    assert x == MyTestClass(
        name="name",
        meta="meta",
        msg=[{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}],
    )



import pytest

from dataclass_wizard.models import Alias


def test_alias_does_not_allow_both_default_and_default_factory():
    """
    Confirm we can't specify both `default` and `default_factory` when
    calling the :func:`Alias` helper function.
    """
    with pytest.raises(ValueError):
        _ = Alias('test', default=None, default_factory=None)



import os
from dataclasses import MISSING
from pathlib import Path

from ..decorators import cached_class_property
from ..lazy_imports import dotenv
from ..utils.string_conv import to_snake_case


# Type of `os.environ` or `DotEnv` dict
Environ = dict[str, 'str | None']

# noinspection PyTypeChecker
environ = None


# noinspection PyMethodParameters
class Env:

    __slots__ = ()

    _accessed_cleaned_to_env = False

    @classmethod
    def load_environ(cls, force_reload=False):
        """
        Load :attr:`environ` from ``os.environ``.

        If `force_reload` is true, start fresh
        and re-copy `os.environ`.
        """
        global environ

        if (_env_not_setup := environ is None) or force_reload:
            # Copy `os.environ`, so as not to mutate it
            environ = os.environ.copy()

            if not _env_not_setup:
                # Refresh `var_names`, in case env variables
                # were removed (deleted) from `os.environ`
                cls.var_names = set(environ)

                if cls._accessed_cleaned_to_env:
                    cls.cleaned_to_env = {
                        k: v for k, v in cls.cleaned_to_env.items()
                        if v in cls.var_names
                    }

    @cached_class_property
    def var_names(cls):
        """
        Cached mapping of `os.environ` key names. This can be refreshed with
        :meth:`reload` as needed.
        """
        return set(environ) if environ is not None else set()

    @classmethod
    def reload(cls, env=None):
        """Refresh cached environment variable names."""
        env_vars = cls.var_names

        if env is None:
            cls.load_environ(force_reload=True)
            env = environ

        new_vars = set(env) - env_vars

        # update names of environment variables
        env_vars.update(new_vars)

        # update mapping of cleaned environment variables (if needed)
        if cls._accessed_cleaned_to_env:
            cls.cleaned_to_env.update(
                (clean(var), var) for var in new_vars
            )

    @classmethod
    def secret_values(cls, dirs):
        """
        Retrieve the values (environment variables) from secret file(s)
        in a secret directory, or a list/tuple of secret directories.
        """
        if isinstance(dirs, (str, os.PathLike)):
            dirs = [dirs]

        env: Environ = {}

        for d in dirs:
            d: Path = d if isinstance(d, os.PathLike) else Path(d)

            if d.exists():
                if d.is_dir():
                    # Iterate over all files in the directory
                    for f in d.iterdir():
                        if f.is_file():  # Ensure it's a file, not a subdirectory
                            env[f.name] = f.read_text()
                elif d.is_file():
                    raise ValueError(f'Secrets directory `{d!r}` is a file, not a directory.')

        return env

    @classmethod
    def update_with_secret_values(cls, dirs):

        secret_values = cls.secret_values(dirs)

        # reload cached mapping of environment variables
        cls.reload(secret_values)
        # update `environ` with new environment variables
        environ.update(secret_values)

    @classmethod
    def dotenv_values(cls, files):
        """
        Retrieve the values (environment variables) from a dotenv file,
        or a list/tuple of dotenv files.
        """
        if isinstance(files, (str, os.PathLike)):
            files = [files]
        elif files is True:
            files = ['.env']

        env: Environ = {}

        for f in files:
            # iterate backwards (from current directory) to find the
            # dotenv file
            dotenv_path = dotenv.find_dotenv(f)
            # take environment variables from `.env` file
            dotenv_values = dotenv.dotenv_values(dotenv_path)
            env.update(dotenv_values)

        return env

    @classmethod
    def update_with_dotenv(cls, files='.env', dotenv_values=None):

        if dotenv_values is None:
            dotenv_values = cls.dotenv_values(files)

        # reload cached mapping of environment variables
        cls.reload(dotenv_values)
        # update `environ` with new environment variables
        environ.update(dotenv_values)

    # noinspection PyDunderSlots,PyUnresolvedReferences,PyClassVar
    @cached_class_property
    def cleaned_to_env(cls):
        cls._accessed_cleaned_to_env = True
        return {clean(var): var for var in cls.var_names}


def clean(s):
    """
    TODO:
        see https://stackoverflow.com/questions/1276764/stripping-everything-but-alphanumeric-chars-from-a-string-in-python
        also, see if we can refactor to use something like Rust and `pyo3` for a slight performance improvement.
    """
    return s.replace('-', '').replace('_', '').lower()


def try_cleaned(key):
    """
    Return the value of the env variable as a *string* if present in
    the Environment, or `MISSING` otherwise.
    """
    key = Env.cleaned_to_env.get(clean(key))

    if key is not None:
        return environ[key]

    return MISSING


if os.name == 'nt':
    # Where Env Var Names Must Be UPPERCASE
    def lookup_exact(var):
        """
        Lookup by variable name(s) with *exact* letter casing, and return
        `None` if not found in the environment.
        """
        if isinstance(var, str):
            var = var.upper()

            if var in Env.var_names:
                return environ[var]

        else:  # a collection of env variable names.
            for v in var:
                v = v.upper()

                if v in Env.var_names:
                    return environ[v]

        return MISSING

else:
    # Where Env Var Names Can Be Mixed Case
    def lookup_exact(var):
        """
        Lookup by variable name(s) with *exact* letter casing, and return
        `None` if not found in the environment.
        """
        if isinstance(var, str):
            if var in Env.var_names:
                return environ[var]

        else:  # a collection of env variable names.
            for v in var:
                if v in Env.var_names:
                    return environ[v]

        return MISSING


def with_screaming_snake_case(field_name):
    """
    Lookup with `SCREAMING_SNAKE_CASE` letter casing first - this is the
    default lookup.

    This function assumes the dataclass field name is lower-cased.

    For a field named 'my_env_var', this tries the following lookups in order:
        - MY_ENV_VAR (screaming snake-case)
        - my_env_var (snake-case)
        - Any other variations - i.e. MyEnvVar, myEnvVar, myenvvar, my-env-var

    :param field_name: The dataclass field name to lookup in the environment.
    :return: The value of the matched environment variable, if one is found in
      the environment.
    """
    upper_key = field_name.upper()

    if upper_key in Env.var_names:
        return environ[upper_key]

    if field_name in Env.var_names:
        return environ[field_name]

    return try_cleaned(field_name)


def with_snake_case(field_name):
    """Lookup with `snake_case` letter casing first.

    This function assumes the dataclass field name is lower-cased.

    For a field named 'my_env_var', this tries the following lookups in order:
        - my_env_var (snake-case)
        - MY_ENV_VAR (screaming snake-case)
        - Any other variations - i.e. MyEnvVar, myEnvVar, myenvvar, my-env-var

    :param field_name: The dataclass field name to lookup in the environment.
    :return: The value of the matched environment variable, if one is found in
      the environment.
    """
    if field_name in Env.var_names:
        return environ[field_name]

    upper_key = field_name.upper()

    if upper_key in Env.var_names:
        return environ[upper_key]

    return try_cleaned(field_name)


def with_pascal_or_camel_case(field_name):
    """Lookup with `PascalCase` or `camelCase` letter casing first.

    This function assumes the dataclass field name is either pascal- or camel-
    cased.

    For a field named 'myEnvVar', this tries the following lookups in order:
        - myEnvVar, MyEnvVar (camel-case, or pascal-case)
        - MY_ENV_VAR (screaming snake-case)
        - my_env_var (snake-case)
        - Any other variations - i.e. my-env-var, myenvvar

    :param field_name: The dataclass field name to lookup in the environment.
    :return: The value of the matched environment variable, if one is found in
      the environment.
    """
    if field_name in Env.var_names:
        return environ[field_name]

    snake_key = to_snake_case(field_name)
    upper_key = snake_key.upper()

    if upper_key in Env.var_names:
        return environ[upper_key]

    if snake_key in Env.var_names:
        return environ[snake_key]

    return try_cleaned(field_name)



import datetime
from collections.abc import Collection
from dataclasses import Field
from types import EllipsisType
from typing import Any, Callable, ClassVar, TypeVar

from _typeshed import Incomplete

from ._bases import AbstractMeta as AbstractMeta
from ._bases import BaseDumpHook as BaseDumpHook
from ._class_helper import (
    dataclass_field_to_skip_if as dataclass_field_to_skip_if,
)
from ._class_helper import (
    resolve_dataclass_field_to_alias_for_dump as resolve_dataclass_field_to_alias_for_dump,
)
from ._class_helper import set_class_dumper as set_class_dumper
from ._decorators import (
    setup_recursive_safe_function as setup_recursive_safe_function,
)
from ._decorators import (
    setup_recursive_safe_function_for_generic as setup_recursive_safe_function_for_generic,
)
from ._meta_cache import create_meta as create_meta
from ._meta_cache import get_meta as get_meta
from ._models import Extras, TypeInfo
from ._type_conv import datetime_to_timestamp as datetime_to_timestamp
from ._type_def import ExplicitNull as ExplicitNull
from ._type_def import JSONObject
from ._type_def import T as T
from ._type_utils import create_new_class as create_new_class
from ._type_utils import is_subclass_safe as is_subclass_safe
from .enums import DateTimeTo as DateTimeTo
from .enums import KeyCase as KeyCase
from .errors import JSONWizardError as JSONWizardError
from .errors import MissingData as MissingData
from .errors import MissingFields as MissingFields
from .errors import ParseError as ParseError
from .utils._dataclass_compat import (
    dataclass_field_names as dataclass_field_names,
)
from .utils._dataclass_compat import dataclass_fields as dataclass_fields
from .utils._dataclass_compat import set_new_attribute as set_new_attribute
from .utils._dict_helper import NestedDict as NestedDict
from .utils._function_builder import FunctionBuilder as FunctionBuilder
from .utils._typing_compat import (
    eval_forward_ref_if_needed as eval_forward_ref_if_needed,
)
from .utils._typing_compat import (
    get_keys_for_typed_dict as get_keys_for_typed_dict,
)
from .utils._typing_compat import get_origin_v2 as get_origin_v2
from .utils._typing_compat import is_annotated as is_annotated
from .utils._typing_compat import is_typed_dict as is_typed_dict
from .utils._typing_compat import (
    is_typed_dict_type_qualifier as is_typed_dict_type_qualifier,
)
from .utils._typing_compat import is_union as is_union

LEAF_TYPES: frozenset
LEAF_TYPES_NO_BYTES: frozenset
ZERO: datetime.timedelta
UTC: datetime.timezone
CLASS_TO_DUMPER: dict
CATCH_ALL: str
TAG: str
PACKAGE_NAME: str
_DUMP_HOOKS: str
_KNOWN_FACTORY_LITERALS: dict
D = TypeVar('D', bound=DumpMixin)

def get_default_dump_hooks(dumper: type[D]) -> dict[type, Callable]: ...
def default_compare_expr(f: Field[Any], locals_ns: dict[str, Any], default_name: str, *, allow_calling_unknown_factories: bool = ...) -> str | None: ...
def _type_returns_value_unchanged(arg, leaf_handling_as_subclass, origin: Incomplete | None = ...): ...
def _all_return_value_unchanged(args, leaf_handling_as_subclass): ...

class DumpMixin(BaseDumpHook):
    transform_dataclass_field: ClassVar[None | EllipsisType] = ...
    __HOOKS__: ClassVar[dict[type, Callable]]
    @classmethod
    def __init_subclass__(cls, _setup_defaults: bool = True, **kwargs): ...
    @staticmethod
    def dump_fallback(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_str(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_int(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_float(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_bool(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_literal(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_bytes(tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_bytearray(cls, tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_none(tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_enum(tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_uuid(tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_iterable(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_tuple(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_named_tuple(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_named_tuple_untyped(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def _build_dict_comp(cls, tp, v, i_next, k_next, v_next, kt, vt, extras): ...
    @classmethod
    def dump_from_dict(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_defaultdict(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_typed_dict(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def _dump_from_typed_dict_fn(cls, _cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_union(cls, _cls, tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_decimal(tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_path(tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_date(cls, tp: TypeInfo, extras: Extras): ...
    @classmethod
    def dump_from_datetime(cls, tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_time(tp: TypeInfo, _extras: Extras): ...
    @staticmethod
    def dump_from_timedelta(tp: TypeInfo, extras: Extras): ...
    @staticmethod
    def dump_from_dataclass(tp: TypeInfo, extras: Extras, _cls: Incomplete | None = ...): ...
    @classmethod
    def dump_dispatcher_for_annotation(cls, tp, extras): ...
def setup_default_dumper(cls: type[DumpMixin] = ...): ...
def check_and_raise_missing_fields(_locals, o, cls, fields: tuple[Field, ...]): ...
def dump_func_for_dataclass(cls: type, extras: Extras | None = ..., dumper_cls: type[DumpMixin] = ..., base_meta_cls: type = ...) -> Callable[[T], JSONObject] | str: ...
def generate_field_code(cls_dumper: DumpMixin, extras: Extras, field: Field, field_i: int, var_name: Incomplete | None = ...) -> str | TypeInfo: ...
def re_raise(e, cls, o, fields, field, value): ...
def get_dumper(class_or_instance: Incomplete | None = ..., create: bool = ..., base_cls: type[D] = ...) -> type[D]: ...
def asdict(o: T, *, cls: Incomplete | None = ..., dict_factory: type[dict] = ..., exclude: Collection[str] | None = ..., **kwargs) -> JSONObject: ...



from __future__ import annotations

import hashlib
from dataclasses import MISSING
from functools import wraps
from typing import TYPE_CHECKING, Callable, cast

from ._type_def import DT
from .utils._function_builder import FunctionBuilder
from .utils._typing_compat import is_union

if TYPE_CHECKING:  # pragma: no cover
    from ._models import Extras, TypeInfo


def process_patterned_date_time(func: Callable) -> Callable:
    """
    Decorator for processing patterned date and time data.

    If the 'pattern' key exists in the `extras` dictionary, it updates
    the base and origin of the type information and processes the
    pattern before calling the original function.

    Supports both class methods and static methods.

    Args:
        func (Callable): The function to decorate, either a class method
        or static method.

    Returns:
        Callable: The wrapped function with pattern processing applied.
    """

    # Determine if the function is a class method
    # noinspection PyUnresolvedReferences
    is_class_method = func.__code__.co_argcount == 3

    if is_class_method:

        @wraps(func)
        def class_method_wrapper(cls, tp: TypeInfo, extras: Extras):
            # Process pattern if it exists in extras
            if (pb := extras.get('pattern')) is not None:
                pb.base = cast(type[DT], tp.origin)
                tp.origin = cast(type, pb)
                return pb.load_to_pattern(tp, extras)

            # Fallback to the original method
            return func(cls, tp, extras)

        return class_method_wrapper
    else:

        @wraps(func)
        def static_method_wrapper(tp: TypeInfo, extras: Extras):
            # Process pattern if it exists in extras
            if (pb := extras.get('pattern')) is not None:
                pb.base = cast(type[DT], tp.origin)
                tp.origin = cast(type, pb)
                return pb.load_to_pattern(tp, extras)

            # Fallback to the original method
            return func(tp, extras)

        return static_method_wrapper


def _type_id(t) -> str:
    # stable-ish identifier for hashing purposes
    mod = getattr(t, '__module__', None)
    qn = getattr(t, '__qualname__', None)
    if mod and qn:
        return f'{mod}.{qn}'
    return repr(t)


def _generic_sig_str(name, args) -> str:
    args = _canonical_union_args(args)  # Union[..]: flattened, de-duped, sorted
    return f'{name}[{",".join(_type_id(a) for a in args)}]'


def _union_args(x):
    # get args similarly to typing.get_args but without importing it everywhere
    return getattr(x, '__args__', ())


def _flatten_union_args(args):
    out = []
    for a in args:
        if is_union(a):
            out.extend(_flatten_union_args(_union_args(a)))
        else:
            out.append(a)
    return out


def _canonical_union_args(args):
    flat = _flatten_union_args(args)
    seen = set()
    uniq = []
    for a in flat:
        k = _type_id(a)
        if k not in seen:
            seen.add(k)
            uniq.append(a)
    uniq.sort(key=_type_id)
    return tuple(uniq)


def setup_recursive_safe_function(
    func: Callable = None,
    *,
    fn_name: str | None = None,
    is_generic: bool = False,
    add_cls: bool = True,
    prefix: str = 'load',
    per_class_cache: bool = False,
) -> Callable:
    """
    A decorator to ensure recursion safety and facilitate dynamic function generation
    with `FunctionBuilder`, supporting both generic and non-generic types.

    The decorated function can define the logic for dynamically generated functions.
    If `fn_name` is provided, the decorator assumes that the function generation
    context (e.g., `with fn_gen.function(...)`) has already been handled externally
    and will not apply it again.

    :param func: The function to decorate. If None, the decorator is applied with arguments.
    :type func: Callable, optional
    :param fn_name: A format string for dynamically generating function names, or None.
    :type fn_name: str, optional
    :param is_generic: Whether the function deals with generic types.
    :type is_generic: bool, optional
    :param add_cls: Whether the class should be added to the function locals
      for `FunctionBuilder`.
    :type add_cls: bool, optional
    :return: The decorated function with recursion safety and dynamic function generation.
    :rtype: Callable
    """

    if func is None:
        return lambda f: setup_recursive_safe_function(
            f,
            fn_name=fn_name,
            is_generic=is_generic,
            add_cls=add_cls,
            prefix=prefix,
            per_class_cache=per_class_cache,
        )

    def _wrapper_logic(tp: TypeInfo, extras: Extras, _cls=None) -> str:
        """
        Shared logic for both class and regular methods. Ensures recursion safety
        and integrates `FunctionBuilder` to dynamically create functions.

        :param tp: The type or generic type being processed.
        :param extras: A context dictionary containing auxiliary information like
                       recursion guards and function builders.
        :type extras: dict
        :param _cls: The class context for class methods. Defaults to None.
        :return: The generated function call expression as a string.
        :rtype: str
        """
        name = tp.name
        if is_generic:
            ann_tp_or_args = (name, _canonical_union_args(tp.args))
        else:
            ann_tp_or_args = tp.origin

        recursion_guard = extras['recursion_guard']

        # new function: drop indices and explicit name
        tp_for_func = tp.replace(index=None, val_name=None)

        if per_class_cache:
            key = (prefix, extras['cls'], ann_tp_or_args)
        else:
            key = (prefix, ann_tp_or_args)

        if (_fn_name := recursion_guard.get(key)) is None:
            cls_name = extras['cls_name']
            tp_name = func.__name__.split('_', 2)[-1]

            # Generate the function name
            if fn_name:
                _fn_name = fn_name.format(cls_name=name)
            else:
                cls_part = f'_{cls_name}' if per_class_cache else ''
                if is_generic:
                    sig_src = _generic_sig_str(name, ann_tp_or_args).encode('utf-8')
                    # noinspection PyTypeChecker
                    sig_hash = hashlib.blake2s(sig_src, digest_size=6).hexdigest()
                    _fn_name = f'_{prefix}{cls_part}_{tp_name}_{sig_hash}'
                else:
                    _fn_name = f'_{prefix}{cls_part}_{tp_name}_{name}'

            recursion_guard[key] = _fn_name

            # Retrieve the main FunctionBuilder
            main_fn_gen = extras['fn_gen']

            # Prepare a new FunctionBuilder for this function
            updated_extras = extras.copy()
            updated_extras['locals'] = _locals = {'cls': ann_tp_or_args} if add_cls else {}
            updated_extras['fn_gen'] = new_fn_gen = FunctionBuilder()

            # Apply the decorated function logic
            if fn_name:
                # Assume `with fn_gen.function(...)` is already handled
                func(_cls, tp_for_func, updated_extras) if _cls else func(tp_for_func, updated_extras)
            else:
                # Apply `with fn_gen.function(...)` explicitly
                with new_fn_gen.function(_fn_name, [tp_for_func.v_for_def()], MISSING, _locals):
                    func(_cls, tp_for_func, updated_extras) if _cls else func(tp_for_func, updated_extras)

            # Merge the new FunctionBuilder into the main one
            main_fn_gen |= new_fn_gen

        return f'{_fn_name}({tp.v()})'

    # Determine if the function is a class method
    # noinspection PyUnresolvedReferences
    is_class_method = func.__code__.co_argcount == 3

    if is_class_method:
        def wrapper_class_method(_cls, tp, extras) -> str:
            """
            Wrapper logic for class methods. Passes the class context to `_wrapper_logic`.

            :param _cls: The class instance.
            :param tp: The type or generic type being processed.
            :param extras: A context dictionary with auxiliary information.
            :type extras: dict
            :return: The generated function call expression as a string.
            :rtype: str
            """
            return _wrapper_logic(tp, extras, _cls)

        wrapper = wraps(func)(wrapper_class_method)
    else:
        wrapper = wraps(func)(_wrapper_logic)

    return wrapper


def setup_recursive_safe_function_for_generic(func: Callable = None,
                                              prefix='load',
                                              per_class_cache: bool = False) -> Callable:
    """
    A helper decorator to handle generic types using
    `setup_recursive_safe_function`.

    Parameters
    ----------
    func : Callable
        The function to be decorated, responsible for returning the
        generated function name.

    Returns
    -------
    Callable
        A wrapped function ensuring recursion safety for generic types.
    """
    return setup_recursive_safe_function(func, is_generic=True, prefix=prefix,
                                         per_class_cache=per_class_cache)


# TODO see if we can remove this
# noinspection PyPep8Naming
class cached_class_property:
    """
    Descriptor decorator implementing a class-level, read-only property,
    which caches the attribute on-demand on the first use.

    Credits: https://stackoverflow.com/a/4037979/10237506
    """
    def __init__(self, func):
        self.__func__ = func
        self.__attr_name__ = func.__name__

    def __get__(self, instance, cls=None):
        """This method is only called the first time, to cache the value."""
        if cls is None:
            cls = type(instance)

        # Build the attribute.
        attr = self.__func__(cls)

        # Cache the value; hide ourselves.
        setattr(cls, self.__attr_name__, attr)

        return attr


class cached_property:
    """
    Descriptor decorator implementing an instance-level, read-only property,
    which caches the attribute on-demand on the first use.
    """
    def __init__(self, func):
        self.__func__ = func
        self.__attr_name__ = func.__name__

    def __get__(self, instance, cls=None):
        """This method is only called the first time, to cache the value."""
        # Build the attribute.
        attr = self.__func__(instance)

        # Cache the value; hide ourselves.
        setattr(instance, self.__attr_name__, attr)

        return attr



# import logging
# from abc import ABC
# from base64 import b64decode
# from collections import deque, defaultdict
# from dataclasses import dataclass, field
# from datetime import datetime, timedelta
# from typing import (Set, FrozenSet, Optional, Union, List,
#                     DefaultDict, Annotated, Literal)
# from uuid import UUID
#
# import pytest
#
# from dataclass_wizard import *
# from dataclass_wizard.class_helper import get_meta
# from dataclass_wizard.constants import TAG
# from dataclass_wizard.errors import ParseError
# from tests.conftest import *
# from tests._typing import *
#
#
# log = logging.getLogger(__name__)
#
#
# def test_asdict_and_fromdict():
#     """
#     Confirm that Meta settings for both `fromdict` and `asdict` are merged
#     as expected.
#     """
#
#     @dataclass
#     class MyClass:
#         my_bool: Optional[bool]
#         myStrOrInt: Union[str, int]
#
#     d = {'myBoolean': 'tRuE', 'my_str_or_int': 123}
#
#     LoadMeta(
#         key_transform='CAMEL',
#         raise_on_unknown_json_key=True,
#         json_key_to_field={'myBoolean': 'my_bool', '__all__': True}
#     ).bind_to(MyClass)
#
#     DumpMeta(key_transform='SNAKE').bind_to(MyClass)
#
#     # Assert that meta is properly merged as expected
#     meta = get_meta(MyClass)
#     assert 'CAMEL' == meta.key_transform_with_load
#     assert 'SNAKE' == meta.key_transform_with_dump
#     assert True is meta.raise_on_unknown_json_key
#     assert {'myBoolean': 'my_bool'} == meta.json_key_to_field
#
#     c = fromdict(MyClass, d)
#
#     assert c.my_bool is True
#     assert isinstance(c.myStrOrInt, int)
#     assert c.myStrOrInt == 123
#
#     new_dict = asdict(c)
#
#     assert new_dict == {'myBoolean': True, 'my_str_or_int': 123}
#
#
# def test_asdict_with_nested_dataclass():
#     """Confirm that `asdict` works for nested dataclasses as well."""
#
#     @dataclass
#     class Container:
#         id: int
#         submittedDt: datetime
#         myElements: List['MyElement']
#
#     @dataclass
#     class MyElement:
#         order_index: Optional[int]
#         status_code: Union[int, str]
#
#     submitted_dt = datetime(2021, 1, 1, 5)
#     elements = [MyElement(111, '200'), MyElement(222, 404)]
#
#     c = Container(123, submitted_dt, myElements=elements)
#
#     DumpMeta(key_transform='SNAKE',
#              marshal_date_time_as='TIMESTAMP').bind_to(Container)
#
#     d = asdict(c)
#
#     expected = {
#         'id': 123,
#         'submitted_dt': round(submitted_dt.timestamp()),
#         'my_elements': [
#             # Key transform now applies recursively to all nested dataclasses
#             # by default! :-)
#             {'order_index': 111, 'status_code': '200'},
#             {'order_index': 222, 'status_code': 404}
#         ]
#     }
#
#     assert d == expected
#
#
# def test_tag_field_is_used_in_dump_process():
#     """
#     Confirm that the `_TAG` field appears in the serialized JSON or dict
#     object (even for nested dataclasses) when a value is set in the
#     `Meta` config for a JSONWizard sub-class.
#     """
#
#     @dataclass
#     class Data(ABC):
#         """ base class for a Member """
#         number: float
#
#     class DataA(Data):
#         """ A type of Data"""
#         pass
#
#     class DataB(Data, JSONWizard):
#         """ Another type of Data """
#         class _(JSONWizard.Meta):
#             """
#             This defines a custom tag that shows up in de-serialized
#             dictionary object.
#             """
#             tag = 'B'
#
#     @dataclass
#     class Container(JSONWizard):
#         """ container holds a subclass of Data """
#         class _(JSONWizard.Meta):
#             tag = 'CONTAINER'
#
#         data: Union[DataA, DataB]
#
#     data_a = DataA(number=1.0)
#     data_b = DataB(number=1.0)
#
#     # initialize container with DataA
#     container = Container(data=data_a)
#
#     # export container to string and load new container from string
#     d1 = container.to_dict()
#
#     expected = {
#         TAG: 'CONTAINER',
#         'data': {'number': 1.0}
#     }
#
#     assert d1 == expected
#
#     # initialize container with DataB
#     container = Container(data=data_b)
#
#     # export container to string and load new container from string
#     d2 = container.to_dict()
#
#     expected = {
#         TAG: 'CONTAINER',
#         'data': {
#             TAG: 'B',
#             'number': 1.0
#         }
#     }
#
#     assert d2 == expected
#
#
# def test_to_dict_key_transform_with_json_field():
#     """
#     Specifying a custom mapping of JSON key to dataclass field, via the
#     `json_field` helper function.
#     """
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         my_str: str = json_field('myCustomStr', all=True)
#         my_bool: bool = json_field(('my_json_bool', 'myTestBool'), all=True)
#
#     value = 'Testing'
#     expected = {'myCustomStr': value, 'my_json_bool': True}
#
#     c = MyClass(my_str=value, my_bool=True)
#
#     result = c.to_dict()
#     log.debug('Parsed object: %r', result)
#
#     assert result == expected
#
#
# def test_to_dict_key_transform_with_json_key():
#     """
#     Specifying a custom mapping of JSON key to dataclass field, via the
#     `json_key` helper function.
#     """
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         my_str: Annotated[str, json_key('myCustomStr', all=True)]
#         my_bool: Annotated[bool, json_key(
#             'my_json_bool', 'myTestBool', all=True)]
#
#     value = 'Testing'
#     expected = {'myCustomStr': value, 'my_json_bool': True}
#
#     c = MyClass(my_str=value, my_bool=True)
#
#     result = c.to_dict()
#     log.debug('Parsed object: %r', result)
#
#     result = c.to_dict()
#     log.debug('Parsed object: %r', result)
#
#     assert result == expected
#
#
# def test_to_dict_with_skip_defaults():
#     """
#     When `skip_defaults` is enabled in the class Meta, fields with default
#     values should be excluded from the serialization process.
#     """
#
#     @dataclass
#     class MyClass(JSONWizard):
#         class _(JSONWizard.Meta):
#             skip_defaults = True
#
#         my_str: str
#         other_str: str = 'any value'
#         optional_str: str = None
#         my_list: List[str] = field(default_factory=list)
#         my_dict: DefaultDict[str, List[float]] = field(
#             default_factory=lambda: defaultdict(list))
#
#     c = MyClass('abc')
#     log.debug('Instance: %r', c)
#
#     out_dict = c.to_dict()
#     assert out_dict == {'myStr': 'abc'}
#
#
# def test_to_dict_with_excluded_fields():
#     """
#     Excluding dataclass fields from the serialization process works
#     as expected.
#     """
#
#     @dataclass
#     class MyClass(JSONWizard):
#
#         my_str: str
#         other_str: Annotated[str, json_key('AnotherStr', dump=False)]
#         my_bool: bool = json_field('TestBool', dump=False)
#         my_int: int = 3
#
#     data = {'MyStr': 'my string',
#             'AnotherStr': 'testing 123',
#             'TestBool': True}
#
#     c = MyClass.from_dict(data)
#     log.debug('Instance: %r', c)
#
#     # dynamically exclude the `my_int` field from serialization
#     additional_exclude = ('my_int', )
#
#     out_dict = c.to_dict(exclude=additional_exclude)
#     assert out_dict == {'myStr': 'my string'}
#
#
# @pytest.mark.parametrize(
#     'input,expected,expectation',
#     [
#         ({1, 2, 3}, [1, 2, 3], does_not_raise()),
#         ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
#     ]
# )
# def test_set(input, expected, expectation):
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         num_set: Set[int]
#         any_set: set
#
#     # Sort expected so the assertions succeed
#     expected = sorted(expected)
#
#     input_set = set(input)
#     c = MyClass(num_set=input_set, any_set=input_set)
#
#     with expectation:
#         result = c.to_dict()
#         log.debug('Parsed object: %r', result)
#
#         assert all(key in result for key in ('numSet', 'anySet'))
#
#         # Set should be converted to list or tuple, as only those are JSON
#         # serializable.
#         assert isinstance(result['numSet'], (list, tuple))
#         assert isinstance(result['anySet'], (list, tuple))
#
#         assert sorted(result['numSet']) == expected
#         assert sorted(result['anySet']) == expected
#
#
# @pytest.mark.parametrize(
#     'input,expected,expectation',
#     [
#         ({1, 2, 3}, [1, 2, 3], does_not_raise()),
#         ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
#     ]
# )
# def test_frozenset(input, expected, expectation):
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         num_set: FrozenSet[int]
#         any_set: frozenset
#
#     # Sort expected so the assertions succeed
#     expected = sorted(expected)
#
#     input_set = frozenset(input)
#     c = MyClass(num_set=input_set, any_set=input_set)
#
#     with expectation:
#         result = c.to_dict()
#         log.debug('Parsed object: %r', result)
#
#         assert all(key in result for key in ('numSet', 'anySet'))
#
#         # Set should be converted to list or tuple, as only those are JSON
#         # serializable.
#         assert isinstance(result['numSet'], (list, tuple))
#         assert isinstance(result['anySet'], (list, tuple))
#
#         assert sorted(result['numSet']) == expected
#         assert sorted(result['anySet']) == expected
#
#
# @pytest.mark.parametrize(
#     'input,expected,expectation',
#     [
#         ({1, 2, 3}, [1, 2, 3], does_not_raise()),
#         ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
#     ]
# )
# def test_deque(input, expected, expectation):
#
#     @dataclass
#     class MyQClass(JSONSerializable):
#         num_deque: deque[int]
#         any_deque: deque
#
#     input_deque = deque(input)
#     c = MyQClass(num_deque=input_deque, any_deque=input_deque)
#
#     with expectation:
#         result = c.to_dict()
#         log.debug('Parsed object: %r', result)
#
#         assert all(key in result for key in ('numDeque', 'anyDeque'))
#
#         # Set should be converted to list or tuple, as only those are JSON
#         # serializable.
#         assert isinstance(result['numDeque'], list)
#         assert isinstance(result['anyDeque'], list)
#
#         assert result['numDeque'] == expected
#         assert result['anyDeque'] == expected
#
#
# @pytest.mark.parametrize(
#     'input,expectation',
#     [
#         ('testing', pytest.raises(ParseError)),
#         ('e1', does_not_raise()),
#         (False, pytest.raises(ParseError)),
#         (0, does_not_raise()),
#     ]
# )
# @pytest.mark.xfail(reason='still need to add the dump hook for this type')
# def test_literal(input, expectation):
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         class Meta(JSONSerializable.Meta):
#             key_transform_with_dump = 'PASCAL'
#
#         my_lit: Literal['e1', 'e2', 0]
#
#     c = MyClass(my_lit=input)
#     expected = {'MyLit': input}
#
#     with expectation:
#         actual = c.to_dict()
#
#         assert actual == expected
#         log.debug('Parsed object: %r', actual)
#
#
# @pytest.mark.parametrize(
#     'input,expectation',
#     [
#         (UUID('12345678-1234-1234-1234-1234567abcde'), does_not_raise()),
#         (UUID('{12345678-1234-5678-1234-567812345678}'), does_not_raise()),
#         (UUID('12345678123456781234567812345678'), does_not_raise()),
#         (UUID('urn:uuid:12345678-1234-5678-1234-567812345678'), does_not_raise()),
#     ]
# )
# def test_uuid(input, expectation):
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         class Meta(JSONSerializable.Meta):
#             key_transform_with_dump = 'Snake'
#
#         my_id: UUID
#
#     c = MyClass(my_id=input)
#     expected = {'my_id': input.hex}
#
#     with expectation:
#         actual = c.to_dict()
#
#         assert actual == expected
#         log.debug('Parsed object: %r', actual)
#
#
# @pytest.mark.parametrize(
#     'input,expectation',
#     [
#         (timedelta(seconds=12345), does_not_raise()),
#         (timedelta(hours=1, minutes=32), does_not_raise()),
#         (timedelta(days=1, minutes=51, seconds=7), does_not_raise()),
#     ]
# )
# def test_timedelta(input, expectation):
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         class Meta(JSONSerializable.Meta):
#             key_transform_with_dump = 'Snake'
#         my_td: timedelta
#
#     c = MyClass(my_td=input)
#     expected = {'my_td': str(input)}
#
#     with expectation:
#         actual = c.to_dict()
#
#         assert actual == expected
#         log.debug('Parsed object: %r', actual)
#
#
# @pytest.mark.parametrize(
#     'input,expectation',
#     [
#         (
#             {}, pytest.raises(ParseError)),
#         (
#             {'key': 'value'}, pytest.raises(ParseError)),
#         (
#             {'my_str': 'test', 'my_int': 2,
#              'my_bool': True, 'other_key': 'testing'}, does_not_raise()),
#         (
#             {'my_str': 3}, pytest.raises(ParseError)),
#         (
#             {'my_str': 'test', 'my_int': 'test', 'my_bool': True},
#             pytest.raises(ValueError)),
#         (
#             {'my_str': 'test', 'my_int': 2, 'my_bool': True},
#             does_not_raise(),
#         )
#     ]
# )
# @pytest.mark.xfail(reason='still need to add the dump hook for this type')
# def test_typed_dict(input, expectation):
#
#     class MyDict(TypedDict):
#         my_str: str
#         my_bool: bool
#         my_int: int
#
#     @dataclass
#     class MyClass(JSONSerializable):
#         my_typed_dict: MyDict
#
#     c = MyClass(my_typed_dict=input)
#
#     with expectation:
#         result = c.to_dict()
#         log.debug('Parsed object: %r', result)
#
#
# def test_using_dataclass_in_dict():
#     """
#     Using dataclass in a dictionary (i.e., dict[str, Test])
#     works as expected.
#
#     See https://github.com/rnag/dataclass-wizard/issues/159
#     """
#     @dataclass
#     class Test:
#         field: str
#
#     @dataclass
#     class Config:
#         tests: dict[str, Test]
#
#     config = {"tests": {"test_a": {"field": "a"}, "test_b": {"field": "b"}}}
#
#     assert fromdict(Config, config) == Config(
#         tests={'test_a': Test(field='a'),
#                'test_b': Test(field='b')})
#
#
# def test_bytes_and_bytes_array_are_supported():
#     """Confirm dump with `bytes` and `bytesarray` is supported."""
#
#     @dataclass
#     class Foo(JSONWizard):
#         b: bytes = None
#         barray: bytearray = None
#         s: str = None
#
#     data = {'b': 'AAAA', 'barray': 'SGVsbG8sIFdvcmxkIQ==', 's': 'foobar'}
#
#     # noinspection PyTypeChecker
#     foo = Foo(b=b64decode('AAAA'),
#               barray=bytearray(b'Hello, World!'),
#               s='foobar')
#
#     # noinspection PyTypeChecker
#     assert foo.to_dict() == data



"""
Dataclass Wizard - a set of wizarding tools for interacting with `dataclasses`
"""

__title__ = 'dataclass-wizard'

__description__ = ('Lightning-fast JSON wizardry for Python dataclasses — '
                   'effortless serialization right out of the box!')
__url__ = 'https://github.com/rnag/dataclass-wizard'
__version__ = '1.0.0'
__author__ = 'Ritvik Nag'
__author_email__ = 'me@ritviknag.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2021-2026 Ritvik Nag'



import os
import sys

# Package name
PACKAGE_NAME = 'dataclass_wizard'

# Library Log Level
LOG_LEVEL = os.getenv('WIZARD_LOG_LEVEL', 'ERROR').upper()

# Current system Python version
_PY_VERSION = sys.version_info[:2]

# Check if currently running Python 3.10 or higher
PY310_OR_ABOVE = _PY_VERSION >= (3, 10)

# Check if currently running Python 3.11 or higher
PY311_OR_ABOVE = _PY_VERSION >= (3, 11)

# Check if currently running Python 3.12 or higher
PY312_OR_ABOVE = _PY_VERSION >= (3, 12)

# Check if currently running Python 3.13 or higher
PY313_OR_ABOVE = _PY_VERSION >= (3, 13)

# Check if currently running Python 3.14 or higher
PY314_OR_ABOVE = _PY_VERSION >= (3, 14)

# The name of the dictionary object that contains `load / dump` hooks for each
# object type. Also used to check if a class is a :class:`BaseLoadHook`
_HOOKS = '__HOOKS__'

# Attribute name that will be defined for single-arg alias functions and
# methods; mainly for internal use.
SINGLE_ARG_ALIAS = '__SINGLE_ARG_ALIAS__'

# Attribute name that will be defined for identity functions and methods;
# mainly for internal use.
IDENTITY = '__IDENTITY__'

# The dictionary key that identifies the tag field for a class. This is only
# set when the `tag` field or the `auto_assign_tags` flag is enabled in the
# `Meta` config for a dataclass.
#
# Note that this key can also be customized in the `Meta` config for a class,
# via the :attr:`tag_key` field.
TAG = '__tag__'

# INTERNAL USE ONLY: The dictionary key that the library
# sets/uses to identify a "catch all" field, which captures
# JSON key/values that don't map to any known dataclass fields.
CATCH_ALL = '<-|CatchAll|->'



from collections import defaultdict
from dataclasses import Field
from typing import Any, Callable, Literal, Sequence, overload

from .abstractions import W, AbstractLoader, AbstractDumper, AbstractParser, E, AbstractLoaderGenerator, AbstractDumperGenerator
from .bases import META, AbstractMeta
from .constants import PACKAGE_NAME
from .models import Condition
from .type_def import ExplicitNullType, T
from .utils.dict_helper import DictWithLowerStore
from .utils.object_path import PathType


# A cached mapping of dataclass to the list of fields, as returned by
# `dataclasses.fields()`.
FIELDS: dict[type, tuple[Field, ...]] = {}

# A cached mapping of dataclass to a mapping of field name
# to default value, as returned by `dataclasses.fields()`.
FIELD_TO_DEFAULT: dict[type, dict[str, Any]] = {}

# Mapping of main dataclass to its `load` function.
CLASS_TO_LOAD_FUNC: dict[type, Any] = {}

# Mapping of main dataclass to its `dump` function.
CLASS_TO_DUMP_FUNC: dict[type, Any] = {}

# A mapping of dataclass to its loader.
CLASS_TO_LOADER: dict[type, type[AbstractLoader]] = {}

# A mapping of dataclass to its dumper.
CLASS_TO_DUMPER: dict[type, type[AbstractDumper]] = {}

# A cached mapping of a dataclass to each of its case-insensitive field names
# and load hook.
FIELD_NAME_TO_LOAD_PARSER: dict[type, DictWithLowerStore[str, AbstractParser]] = {}

# Since the dump process doesn't use Parsers currently, we use a sentinel
# mapping to confirm if we need to setup the dump config for a dataclass
# on an initial run.
IS_DUMP_CONFIG_SETUP: dict[type, bool] = {}

# A cached mapping, per dataclass, of JSON field to instance field name
JSON_FIELD_TO_DATACLASS_FIELD: dict[type, dict[str, str | ExplicitNullType]] = defaultdict(dict)

# A cached mapping, per dataclass, of instance field name to JSON path
DATACLASS_FIELD_TO_JSON_PATH: dict[type, dict[str, PathType]] = defaultdict(dict)

# A cached mapping, per dataclass, of instance field name to alias
DATACLASS_FIELD_TO_ALIAS: dict[type, dict[str, str]] = defaultdict(dict)

# A cached mapping, per dataclass, of instance field name to `SkipIf` condition
DATACLASS_FIELD_TO_SKIP_IF: dict[type, dict[str, Condition]] = defaultdict(dict)

# A cached mapping, per `EnvWizard` subclass, of field name to env variable
FIELD_TO_ENV_VAR: dict[type, dict[str, str]] = defaultdict(dict)

# A mapping of dataclass name to its Meta initializer (defined in
# :class:`bases.BaseJSONWizardMeta`), which is only set when the
# :class:`JSONSerializable.Meta` is sub-classed.
META_INITIALIZER: dict[str, Callable[[type[W]], None]] = {}

# Mapping of dataclass to its Meta inner class, which will only be set when
# the :class:`JSONSerializable.Meta` is sub-classed.
_META: dict[type, META] = {}


def dataclass_to_dumper(cls: type) -> type[AbstractDumper]:
    """
    Returns the dumper for a dataclass.
    """


def set_class_loader(cls_to_loader, class_or_instance, loader: type[AbstractLoader]):
    """
    Set (and return) the loader for a dataclass.
    """


def set_class_dumper(cls: type, dumper: type[AbstractDumper]):
    """
    Set (and return) the dumper for a dataclass.
    """


def json_field_to_dataclass_field(cls: type) -> dict[str, str | ExplicitNullType]:
    """
    Returns a mapping of JSON field to dataclass field.
    """


def dataclass_field_to_json_path(cls: type) -> dict[str, PathType]:
    """
    Returns a mapping of dataclass field to JSON path.
    """


def dataclass_field_to_json_field(cls: type) -> dict[str, str]:
    """
    Returns a mapping of dataclass field to JSON field.
    """


def dataclass_field_to_skip_if(cls: type) -> dict[str, Condition]:
    """
    Returns a mapping of dataclass field to SkipIf condition.
    """


def field_to_env_var(cls: type[E]) -> dict[str, str]:
    """
    Returns a mapping of field in the `EnvWizard` subclass to env variable.
    """


def dataclass_field_to_load_parser(
        cls_loader: type[AbstractLoader],
        cls: type,
        config: META,
        save: bool = True) -> DictWithLowerStore[str, AbstractParser]:
    """
    Returns a mapping of each lower-cased field name to its annotated type.
    """


def _setup_load_config_for_cls(cls_loader: type[AbstractLoader],
                               cls: type,
                               config: META,
                               save: bool = True
                               ) -> DictWithLowerStore[str, AbstractParser]:
    """
    This function processes a class `cls` on an initial run, and sets up the
    load process for `cls` by iterating over each dataclass field. For each
    field, it performs the following tasks:

        * Lookup the Parser (dispatcher) for the field based on its type
          annotation, and then cache it so we don't need to lookup each time.

        * Check if the field's annotation is of type ``Annotated``. If so,
          we iterate over each ``Annotated`` argument and find any special
          :class:`JSON` objects (this can also be set via the helper function
          ``json_key``). Assuming we find it, the class-specific mapping of
          JSON key to dataclass field name is then updated with the input
          passed in to this object.

        * Check if the field type is a :class:`JSONField` object (this can
          also be set by the helper function ``json_field``). Assuming this is
          the case, the class-specific mapping of JSON key to dataclass field
          name is then updated with the input passed in to the :class:`JSON`
          attribute.
    """


def setup_dump_config_for_cls_if_needed(cls: type) -> None:
    """
    This function processes a class `cls` on an initial run, and sets up the
    dump process for `cls` by iterating over each dataclass field. For each
    field, it performs the following tasks:

        * Check if the field's annotation is of type ``Annotated``. If so,
          we iterate over each ``Annotated`` argument and find any special
          :class:`JSON` objects (this can also be set via the helper function
          ``json_key``). Assuming we find it, the class-specific mapping of
          dataclass field name to JSON key is then updated with the input
          passed in to this object.

        * Check if the field type is a :class:`JSONField` object (this can
          also be set by the helper function ``json_field``). Assuming this is
          the case, the class-specific mapping of dataclass field name to JSON
          key is then updated with the input passed in to the :class:`JSON`
          attribute.
    """


def call_meta_initializer_if_needed(cls: type[W | E],
                                    package_name=PACKAGE_NAME) -> None:
    """
    Calls the Meta initializer when the inner :class:`Meta` is sub-classed.
    """


def get_meta(cls: type, base_cls: T = AbstractMeta) -> T | META:
    """
    Retrieves the Meta config for the :class:`AbstractJSONWizard` subclass.

    This config is set when the inner :class:`Meta` is sub-classed.
    """


def create_meta(cls: type, cls_name: str | None = None, **kwargs) -> None:
    """
    Sets the Meta config for the :class:`AbstractJSONWizard` subclass.

    WARNING: Only use if the Meta config is undefined,
      e.g. `get_meta` for the `cls` returns `base_cls`.

    """


def dataclass_fields(cls: type) -> tuple[Field, ...]:
    """
    Cache the `dataclasses.fields()` call for each class, as overall that
    ends up around 5x faster than making a fresh call each time.

    """

@overload
def dataclass_init_fields(cls: type, as_list: Literal[True] = False) -> list[Field]:
    """Get only the dataclass fields that would be passed into the constructor."""


@overload
def dataclass_init_fields(cls: type, as_list: Literal[False] = False) -> tuple[Field]:
    """Get only the dataclass fields that would be passed into the constructor."""


def dataclass_field_names(cls: type) -> tuple[str, ...]:
    """Get the names of all dataclass fields"""


def dataclass_init_field_names(cls: type) -> tuple[str, ...]:
    """Get the names of all __init__() dataclass fields"""


def dataclass_kw_only_init_field_names(cls: type) -> set[str]:
    """Get the names of all "KEYWORD-ONLY" dataclass fields"""


def dataclass_field_to_default(cls: type) -> dict[str, Any]:
    """Get default values for the (optional) dataclass fields."""


def is_builtin(o: Any) -> bool:
    """Check if an object/singleton/class is a builtin in Python."""


def create_new_class(
        class_or_instance, bases: tuple[T, ...],
        suffix: str | None = None, attr_dict=None) -> T:
    """
    Create (dynamically) and return a new class that sub-classes from a list
    of `bases`.
    """


def get_class_name(class_or_instance) -> str:
    """Return the fully qualified name of a class."""


def get_outer_class_name(inner_cls, default=None, raise_: bool = True) -> str:
    """
    Attempt to return the fully qualified name of the outer (enclosing) class,
    given a reference to the inner class.

    If any errors occur - such as when `inner_cls` is not a real inner
    class - then an error will be raised if `raise_` is true, and if not
    we will return `default` instead.

    """


def get_class(obj: Any) -> type:
    """Get the class for an object `obj`"""


def is_subclass(obj: Any, base_cls: type) -> bool:
    """Check if `obj` is a sub-class of `base_cls`"""


def is_subclass_safe(cls, class_or_tuple) -> bool:
    """Check if `obj` is a sub-class of `base_cls` (safer version)"""



__all__ = [
    # Base exports
    'DataclassWizard',
    'JSONWizard',
    'EnvWizard',
    # Helper functions
    'asdict',
    'fromdict',
    'fromlist',
    'register_type',
    'LoadMeta',
    'DumpMeta',
    'EnvMeta',
    # Models
    'Alias',
    'AliasPath',
    'Env',
    'skip_if_field',
]

from ._bases_meta import register_type
from ._dumpers import asdict
from ._loaders import fromdict, fromlist
from ._serial_json import DataclassWizard, JSONWizard
from .env import EnvWizard
from .meta import DumpMeta, EnvMeta, LoadMeta
from .models import Alias, AliasPath, Env, skip_if_field



"""
Generates a Python (dataclass) schema, given a JSON input. The entry point for
this module is the `gen-schema` subcommand.

This JSON to Dataclass conversion tool was inspired by the following projects:

    * https://github.com/mischareitsma/json2dataclass
    * https://github.com/russbiggs/json2dataclass
    * https://github.com/mholt/json-to-go

The parser supports the full JSON spec, so both `list` and `dict` as the
root type are properly handled as expected.

A few important notes on the behavior of JSON parsing:

    * Lists with multiple dictionaries will have all the keys and type
      definitions merged into a single model dataclass, as the dictionary
      objects are considered homogenous in this case.

    * Nested lists within the above structure (e.g. list -> dict -> list)
      should similarly merge all list elements with the list for that same key
      in each sibling `dict` object. For example, assuming the below input::
        ... [{"d1": [1, {"k": "v"}]}, {"d1": [{"k": 2}, {"k2": "v2"}, True]}]
      This should result in a single, merged type definition for "d1"::
        ... List[Union[int, dataclass(k: Union[str, int], k2: str), bool]]

    * Any nested dictionaries within lists will have their Model class name
      generated with the singular form of the key containing the model
      definition -- for example, {"Items":[{"key":"value"}]} will result in a
      model class named `Item`. In the case a dictionary is nested within a
      list, it will have the class name auto-incremented with a common
      prefix -- for example, `Data1`, `Data2`, etc.


The implementation below uses regex code in the `rules.english` module from
the library Python-Inflector (https://github.com/bermi/Python-Inflector).

This library is available under the BSD license, which can be
obtained from https://opensource.org/licenses.

The library Python-Inflector contains the following attribution notices:

    Copyright (c) 2006 Bermi Ferrer Martinez
    bermi a-t bermilabs - com

See the end of this file for the original BSD-style license from this library.

"""

__all__ = [
    'PyCodeGenerator'
]

import json
import re
import textwrap
from collections import defaultdict
from collections import deque
from collections.abc import Iterable
from dataclasses import dataclass, field, InitVar
from datetime import date, datetime, time
from enum import Enum
from pathlib import Path
from typing import Callable, Any, Optional, TypeVar, Type, ClassVar
from typing import DefaultDict, Set, List
from typing import (
    Union, Dict, Sequence
)

from .. import property_wizard
from ..constants import PACKAGE_NAME
from ..class_helper import get_class_name
from ..type_def import PyDeque, JSONList, JSONObject, JSONValue, T
from ..utils.string_conv import to_snake_case, to_pascal_case
# noinspection PyProtectedMember
from ..utils.type_conv import TRUTHY_VALUES
from ..utils.type_conv import as_datetime, as_date, as_time


# Some unconstrained type variables.  These are used by the container types.
# (These are not for export.)
_S = TypeVar('_S')

# Merge both the "truthy" and "falsy" values, so we can determine the criteria
# under which a string can be considered as a boolean value.
_FALSY_VALUES = {'false', 'f', 'no', 'n', 'off', '0'}
_BOOL_VALUES = TRUTHY_VALUES | _FALSY_VALUES

# Valid types for JSON contents; this can be either a list of any type,
# or a dictionary with `string` keys and values of any type.
JSONBlobType = Union[JSONList, JSONObject]

PyDataTypeOrSeq = Union['PyDataType', Sequence['PyDataType']]
TypeContainerElements = Union[PyDataTypeOrSeq,
                              'PyDataclassGenerator', 'PyListGenerator']


@dataclass
class PyCodeGenerator:
    """
    This is the main class responsible for generating Python code that
    leverages dataclasses, given a JSON object as an input data.
    """

    # Either the file name (ex. file1.json) or the file contents as a string
    # can be passed in as an input to the constructor method.
    file_name: InitVar[str] = None
    file_contents: InitVar[str] = None

    # Should we force-resolve inferred types for strings? For example, a value
    # of "TRUE" will appear as a `Union[str, bool]` type by default.
    force_strings: InitVar[bool] = None

    # Enable experimental features via a `__future__` import, which allows
    # PEP-585 and PEP-604 style annotations in Python 3.7+
    experimental: InitVar[bool] = None

    # The rest of these fields are just for internal use.
    parser: 'JSONRootParser' = field(init=False)
    data: JSONBlobType = field(init=False)
    _py_code_lines: List[str] = field(default=None, init=False)

    def __post_init__(self, file_name: str, file_contents: str,
                      force_strings: bool, experimental: bool):

        # Set global flags
        global Globals
        Globals = _Globals(force_strings=force_strings,
                           experimental=experimental)

        # https://stackoverflow.com/a/62940588/10237506
        if file_name:
            file_path = Path(file_name)
            file_contents = file_path.read_bytes()

        self.data = json.loads(file_contents)
        self.parser = JSONRootParser(self.data)

    @property
    def py_code(self) -> str:

        if self._py_code_lines is None:
            # Generate Python code for the dataclass(es)
            dataclass_code: str = repr(self.parser)
            # Add any imports used at the top of the code
            self._py_code_lines = ModuleImporter.imports
            if self._py_code_lines:
                self._py_code_lines.append('')
            # Generate final Python code - imports + dataclass(es)
            self._py_code_lines.append(dataclass_code)

        return '\n'.join(self._py_code_lines)


# Global flags (generally passed in via command-line) which are shared by
# classes and functions.
Globals: '_Globals | None' = None


@dataclass
class _Globals:

    # Should we force-resolve inferred types for strings? For example, a value
    # of "TRUE" will appear as a `Union[str, bool]` type by default.
    force_strings: bool = False

    # Enable experimental features via a `__future__` import, which allows
    # PEP-585 and PEP-604 style annotations in Python 3.7+
    experimental: bool = False

    # Should we insert auto-generated comments under each dataclass.
    insert_comments: bool = True

    # Should we include a newline after the comments block mentioned above.
    newline_after_class_def: bool = True


# Credits: https://github.com/bermi/Python-Inflector
class English:
    """
    Inflector for pluralize and singularize English nouns.

    This is the default Inflector for the Inflector obj
    """

    @staticmethod
    def humanize(word):
        """
        Returns a human-readable string from word, by replacing
        underscores with a space, and by upper-casing the initial
        character by default.
        """
        return to_snake_case(word).replace('_', ' ').title()

    @staticmethod
    def singularize(word):
        """Singularizes English nouns."""

        rules = [
            ['(?i)(quiz)zes$', '\\1'],
            ['(?i)(matr)ices$', '\\1ix'],
            ['(?i)(vert|ind)ices$', '\\1ex'],
            ['(?i)^(ox)en', '\\1'],
            ['(?i)(alias|status)es$', '\\1'],
            ['(?i)([octop|vir])i$', '\\1us'],
            ['(?i)(cris|ax|test)es$', '\\1is'],
            ['(?i)(shoe)s$', '\\1'],
            ['(?i)(o)es$', '\\1'],
            ['(?i)(bus)es$', '\\1'],
            ['(?i)([m|l])ice$', '\\1ouse'],
            ['(?i)(x|ch|ss|sh)es$', '\\1'],
            ['(?i)(m)ovies$', '\\1ovie'],
            ['(?i)(s)eries$', '\\1eries'],
            ['(?i)([^aeiouy]|qu)ies$', '\\1y'],
            ['(?i)([lr])ves$', '\\1f'],
            ['(?i)(tive)s$', '\\1'],
            ['(?i)(hive)s$', '\\1'],
            ['(?i)([^f])ves$', '\\1fe'],
            ['(?i)(^analy)ses$', '\\1sis'],
            ['(?i)(^analysis)$', '\\1'],
            ['(?i)((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$', '\\1\\2sis'],
            # I don't want 'Data' replaced with 'Datum', however
            ['(?i)(^data)$', '\\1'],
            ['(?i)([ti])a$', '\\1um'],
            ['(?i)(n)ews$', '\\1ews'],
            ['(?i)s$', ''],
        ]

        uncountable_words = ['equipment', 'information', 'rice', 'money',
                             'species', 'series', 'fish', 'sheep', 'sms']

        irregular_words = {
            'people': 'person',
            'men': 'man',
            'children': 'child',
            'sexes': 'sex',
            'moves': 'move'
        }

        lower_cased_word = word.lower()

        for uncountable_word in uncountable_words:
            if lower_cased_word[-1 * len(uncountable_word):] == uncountable_word:
                return word

        for irregular in irregular_words.keys():
            match = re.search('(' + irregular + ')$', word, re.IGNORECASE)
            if match:
                return re.sub(
                    '(?i)' + irregular + '$',
                    match.expand('\\1')[0] + irregular_words[irregular][1:],
                    word)

        for rule in range(len(rules)):
            match = re.search(rules[rule][0], word, re.IGNORECASE)
            if match:
                groups = match.groups()
                for k in range(0, len(groups)):
                    if groups[k] == None:
                        rules[rule][1] = rules[
                            rule][1].replace('\\' + str(k + 1), '')

                return re.sub(rules[rule][0], rules[rule][1], word)

        return word


# noinspection SpellCheckingInspection, PyPep8Naming
class classproperty:
    """
    Decorator that converts a method with a single cls argument into a
    property that can be accessed directly from the class.

    Credits:
        - https://stackoverflow.com/a/57055258/10237506
        - https://docs.djangoproject.com/en/3.1/ref/utils/#django.utils.functional.classproperty

    """
    def __init__(self, method: Callable[[Any], T]) -> None:
        self.f = method

    def __get__(
            self, instance: Optional[_S], cls: Optional[Type[_S]] = None) -> T:
        return self.f(cls)

    def getter(self, method):
        self.f = method
        return self


def is_float(s: str) -> bool:
    """
    Check if a string is a :class:`float` value
      ex. '1.23'
    """
    try:
        _ = float(s)
        return True
    except ValueError:
        return False


def can_be_bool(o: str) -> bool:
    """
    Check if a string can be a :class:`bool` value. Note this doesn't mean
    that the string can or should be converted to bool, only that it *appears*
    to be one.

    """
    return o.lower() in _BOOL_VALUES


class PyDataType(Enum):
    """
    Enum representing a Python Data Type
    """
    STRING = str
    FLOAT = float
    INT = int
    BOOL = bool
    LIST = list
    DICT = dict
    DATE = date
    DATETIME = datetime
    TIME = time
    NULL = None

    def __str__(self) -> str:
        """
        Returns the string representation of an Enum member's value.
        """
        return getattr(
            self.value, '__name__', str(self.value))


class ModuleImporter:
    """
    Helper class responsible for constructing import statements in the
    generated Python code.
    """

    # Import level (e.g. stdlib or 3rd party) -> Module Name -> Module Imports
    _MOD_IMPORTS: DefaultDict[int, DefaultDict[str, Set[str]]] = defaultdict(
        lambda: defaultdict(set)
    )

    # noinspection PyMethodParameters
    @classproperty
    def imports(cls: Type[T]) -> List[str]:
        """
        Returns a list of generated import statements based on the modules
        currently used in the code.
        """

        lines = []

        for lvl in sorted(cls._MOD_IMPORTS):
            modules = cls._MOD_IMPORTS[lvl]
            for mod in sorted(modules):
                imported = sorted(modules[mod])
                lines.append(f'from {mod} import {", ".join(imported)}')
            lines.append('')

        return lines

    @classmethod
    def wrap_string_with_import(cls, string: str,
                                imported: object,
                                wrap_chars='[]',
                                register_import=True,
                                level=1) -> str:
        """
        Wraps `string` so it is contained within `imported`. The `wrap_chars`
        parameter determines the enclosing characters to use -- defaults to
        braces by default, as subscripted type Generics often appear in this
        form.

        If `register_import` is true (default), an import statement will also
        be generated for the `imported` usage, if one needs to be added.

        Examples::

            >>> ModuleImporter.wrap_string_with_import('int', List)
            'List[int]'

        """

        module = imported.__module__
        name = cls._get_import_name(imported)
        start, end = wrap_chars

        if register_import:
            cls.register_import_by_name(module, name, level)

        return f'{name}{start}{string}{end}'

    # noinspection PyUnresolvedReferences
    @classmethod
    def wrap_with_import(cls, deck: PyDeque[str],
                         imported: object,
                         wrap_chars='[]',
                         register_import=True,
                         level=1) -> None:
        """
        Same as :meth:`wrap_string_with_import` above, except this accepts
        a list (deque) of strings to be wrapped instead.
        """

        module = imported.__module__
        name = cls._get_import_name(imported)
        start, end = wrap_chars

        if register_import:
            cls.register_import_by_name(module, name, level)

        deck.appendleft(start)
        deck.appendleft(name)
        deck.append(end)

    @classmethod
    def register_import(cls, imported: object, level=1) -> None:
        """
        Registers a new import for the given object.

        Examples::

            >>> ModuleImporter.register_import(datetime)

        """

        module = imported.__module__
        name = cls._get_import_name(imported)

        cls.register_import_by_name(module, name, level)

    @classmethod
    def register_import_by_name(cls, module: str, name: str, level: int) -> None:
        """
        Registers a new import for a module and the imported name.

        Note: any built-in's like "int" or "min" should be skipped by
        default.
        """

        # Skip any built-in helper functions
        #   if name in __builtins__.__dict__:
        if module == 'builtins':
            return

        cls._MOD_IMPORTS[level][module].add(name)

    @classmethod
    def register_future_import(cls, name: str) -> None:
        """
        Registers a top-level `__future__` import for a module, which is
        required to be the first import defined at the top of the file.

        """
        cls._MOD_IMPORTS[0]['__future__'].add(name)

    @classmethod
    def clear_imports(cls):
        """
        Clears all the module imports currently in the cache.
        """

        cls._MOD_IMPORTS.clear()

    @classmethod
    def _get_import_name(cls, imported: Any) -> str:
        """Retrieves the name of an imported object."""
        return cls._safe_get_class_name(imported)

    @staticmethod
    def _safe_get_class_name(cls: Any):
        """
        Retrieves the class name of the specified object or class.

        Note: the `_name` attribute is specific to most Generic types in
        the `typing` module.
        """

        try:
            return cls._name

        except AttributeError:
            # Useful to strip underscores from the start, for example
            # in Python 3.6 which doesn't have a `_name` attribute for the
            # `Union` type, and the class name is returned as `_Union`.
            return get_class_name(cls).lstrip('_')


@dataclass(repr=False)
class TypeContainer(List[TypeContainerElements]):
    """
    Custom list class which functions as a container for Python data types.
    """

    # This keeps track of whether we've seen a `null` type before.
    is_optional = False

    def append(self, o: TypeContainerElements):
        """
        Appends an object (or a sequence of objects) to the
        :class:`TypeContainer` instance.
        """

        if isinstance(o, Iterable):
            for elem in o:
                self.append(elem)
            return

        if o is PyDataType.NULL:
            self.is_optional = True
            return

        if o in self:
            return

        if isinstance(o, PyDataType):
            # Register the types in case they are not standard imports.
            # For example, `uuid` and `datetime` objects.
            ModuleImporter.register_import(o.value)

        super(TypeContainer, self).append(o)

    def __or__(self, other):
        """
        Performs logical OR, to merge instances of :class:`TypeContainer`
        """

        if not isinstance(other, TypeContainer):
            raise TypeError(
                f'TypeContainer: incorrect type for __add__: {type(other)}')

        # Remember to carry over the `is_optional` flag
        self.is_optional |= other.is_optional

        if len(self) == 1 and len(other) == 1:
            self_item = self[0]
            other_item = other[0]

            for typ in PyDataclassGenerator, PyListGenerator:
                if isinstance(self_item, typ) and isinstance(other_item, typ):
                    # We call  `__or__` to merge the lists or dataclasses
                    # together.
                    self_item |= other_item

                    return self

        for elem in other:
            self.append(elem)

        return self

    def __repr__(self):
        """
        Iteratively calls the `repr` method of all our model collection types.
        """

        lines = []

        for typ in self:
            if isinstance(typ, (PyDataclassGenerator, PyListGenerator)):
                lines.append(repr(typ))

        return '\n'.join(lines)

    def __str__(self):
        ...

    def _default_str(self):
        """
        Return the string representation of the resolved type -
          ex.`Optional[Union[str, int]]`

        """

        # I'm using `deque`s here to avoid doing `list.insert(0, x)` or later
        # iterating over `reversed(list)`, as this might be a bit faster.
        # noinspection PyUnresolvedReferences
        typing_imports: PyDeque[object] = deque()
        # noinspection PyUnresolvedReferences
        parts: PyDeque[str]

        if not self:
            # This is the case when the only value encountered for a field is
            # a `null` - hence, we're unable to determine the type.
            typing_imports.appendleft(Any)

        elif self.is_optional:
            typing_imports.appendleft(Optional)

        if len(self) > 1:
            # Else, if we have more than one type for a field, then the
            # resolved type should be a `Union` of all the seen types.
            typing_imports.appendleft(Union)

        parts = deque(', '.join(str(typ) for typ in self))

        for tp in typing_imports:
            ModuleImporter.wrap_with_import(parts, tp)

        return ''.join(parts).replace('[]', '')

    def _experimental_features_str(self):

        if not self:
            # This is the case when the only value encountered for a field is
            # a `null` - hence, we're unable to determine the type.
            ModuleImporter.register_import(Any)
            return 'Any'

        parts = [str(typ) for typ in self]
        if self.is_optional:
            parts.append('None')

        return ' | '.join(parts)


def possible_types_for_string_value(string: str) -> PyDataTypeOrSeq:
    """
    Returns possible types for a JSON field with a :class:`string` value,
    depending on what that value appears to be.

    If `Globals.force_strings` is true and there is more than one possible
    type, we simply return the inferred type, instead of the
    `Union[T..., str]` syntax.
    """

    exc_types = TypeError, ValueError

    try:
        _ = as_date(string)
        return PyDataType.DATE
    except exc_types:
        pass

    # I want to eliminate false positives so this seems the easiest
    # way to do that. Otherwise strings like "24" seem to get parsed
    # as a :class:`Time` object, which might not be expected.
    if ':' not in string:
        possible_types = []

        if string.isnumeric():
            possible_types.append(PyDataType.INT)

        elif is_float(string):
            possible_types.append(PyDataType.FLOAT)

        elif can_be_bool(string):
            possible_types.append(PyDataType.BOOL)

        # If force-resolve is enabled, just return the inferred type if one
        # was determined.
        # noinspection PyUnresolvedReferences
        if Globals.force_strings and possible_types:
            return possible_types[0]

        possible_types.append(PyDataType.STRING)

        return possible_types

    try:
        _ = as_time(string)
        return PyDataType.TIME
    except exc_types:
        pass

    try:
        _ = as_datetime(string)
        return PyDataType.DATETIME
    except exc_types:
        pass

    return PyDataType.STRING


def json_to_python_type(o: JSONValue) -> PyDataTypeOrSeq:
    """
    Convert a JSON object to a Python Data Type, or a Union of Python Data
    Types.
    """

    if o is None:
        return PyDataType.NULL

    if isinstance(o, str):
        return possible_types_for_string_value(o)

    # `bool` needs to come before `int`, as it's a subclass of `int`
    if isinstance(o, bool):
        return PyDataType.BOOL

    if isinstance(o, int):
        return PyDataType.INT

    if isinstance(o, float):
        return PyDataType.FLOAT

    if isinstance(o, list):
        return PyDataType.LIST

    if isinstance(o, dict):
        return PyDataType.DICT


@dataclass
class JSONRootParser:

    data: JSONBlobType

    model: Union['PyListGenerator',
                 'PyDataclassGenerator'] = field(init=False)

    def __post_init__(self):

        # Clear imports from last run
        ModuleImporter.clear_imports()

        str_method_prefix = 'default'

        # Check if experimental features are enabled
        if Globals.experimental:
            # Add the required `__future__` import
            ModuleImporter.register_future_import('annotations')
            # Update how annotations are resolved
            str_method_prefix = 'experimental_features'

        # Set the `__str__` method to use for classes
        str_method_name = f'_{str_method_prefix}_str'
        for typ in TypeContainer, PyListGenerator, PyDataclassGenerator:
            typ.__str__ = getattr(typ, str_method_name)

        # We'll need an import for the @dataclass decorator, at a minimum
        ModuleImporter.register_import(dataclass)

        if isinstance(self.data, list):
            self.model = PyListGenerator(self.data,
                                         is_root=True)

        elif isinstance(self.data, dict):
            self.model = PyDataclassGenerator(self.data,
                                              is_root=True)

        else:
            raise TypeError(
                'Incorrect type, expected a JSON `list` or `dict`. '
                f'actual_type={type(self.data)!r}, data={self.data!r}')

    def __repr__(self):
        return repr(self.model) + '\n'


@dataclass
class PyDataclassGenerator(metaclass=property_wizard):

    data: InitVar[JSONObject]

    _name: str = 'data'
    indent: str = ' ' * 4
    is_root: bool = False

    nested_lvl: InitVar[int] = 0

    parsed_types: DefaultDict[str, TypeContainer] = field(
        init=False,
        default_factory=lambda: defaultdict(TypeContainer)
    )

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        """Title case the name"""
        self._name = to_pascal_case(name)

    @classmethod
    def load_parsed(
            cls: Type[T],
            parsed_types: Dict[str,
                               Union[PyDataType, 'PyDataclassGenerator']],
            **constructor_kwargs
    ) -> T:

        obj = cls({}, **constructor_kwargs)

        for k, typ in parsed_types.items():
            underscored_field = to_snake_case(k)
            obj.parsed_types[underscored_field].append(typ)

        return obj

    def __post_init__(self, data: JSONObject, nested_lvl: int):

        for k, v in data.items():
            underscored_field = to_snake_case(k)
            typ = json_to_python_type(v)

            if typ is PyDataType.DICT:
                typ = PyDataclassGenerator(
                    v, k,
                    nested_lvl=nested_lvl,
                )
            elif typ is PyDataType.LIST:
                nested_lvl += 1
                typ = PyListGenerator(
                    v, k, k,
                    nested_lvl=nested_lvl,
                )

            self.parsed_types[underscored_field].append(typ)

    def __or__(self, other):
        if not isinstance(other, PyDataclassGenerator):
            raise TypeError(
                f'{self.__class__.__name__}: Incorrect type for `__or__`. '
                f'actual_type: {type(other)}, object={other}')

        for k, v in other.parsed_types.items():
            if k in self.parsed_types:
                self.parsed_types[k] |= v

            else:
                self.parsed_types[k] = v

        return self

    def get_lines(self) -> List[str]:
        if self.is_root:
            ModuleImporter.register_import_by_name(
                PACKAGE_NAME, 'JSONWizard', level=2)
            class_name = f'class {self.name}(JSONWizard):'
        else:
            class_name = f'class {self.name}:'

        class_parts = ['@dataclass',
                       class_name]
        parts = []
        nested_parts = []

        # noinspection PyUnresolvedReferences
        if Globals.insert_comments:
            class_parts.append(
                textwrap.indent('"""', self.indent))
            class_parts.append(
                textwrap.indent(f'{self.name} dataclass', self.indent))

            # noinspection PyUnresolvedReferences
            if Globals.newline_after_class_def:
                class_parts.append('')

            class_parts.append(textwrap.indent(
                '"""', self.indent))

        for k, v in self.parsed_types.items():
            line = f'{k}: {v}'
            wrapped_line = textwrap.indent(line, self.indent)
            parts.append(wrapped_line)

            nested_part = repr(v)
            if nested_part:
                nested_parts.append(nested_part)

        for part in nested_parts:
            parts.append('\n')
            parts.append(part)

        if not parts:
            parts = [textwrap.indent('pass', self.indent)]

        class_parts.extend(parts)

        return class_parts

    def __str__(self):
        ...

    def _default_str(self):
        return f"'{self.name}'"

    def _experimental_features_str(self):
        return self.name

    def __repr__(self):
        """
        Returns the Python `dataclasses` representation of the object.
        """
        return '\n'.join(self.get_lines())


@dataclass(repr=False)
class PyListGenerator(metaclass=property_wizard):
    """
    Parse a list in a JSON object to a Python list, based on the following
    rules:

    * If the JSON list contains *only* simple types, for example int,
      str, or bool, then invoking ``str()`` on this object should return
      a Union representation of those types, for example
      `Union[int, str, bool]`.

    * If the JSON list contains *any* complex type, like a dict, then
      all `dict`s should have their keys and values merged together.
      Optional and Union should be included if needed.

      Additionally, if `is_root` is true, then calling ``str()`` will
      effectively ignore any simple types,

    """

    # Default name for model class if none is provided.
    default_name: ClassVar[str] = 'data'

    data: JSONList

    container_name: str = 'container'
    _name: str = None

    indent: str = ' ' * 4

    is_root: InitVar[bool] = False
    nested_lvl: InitVar[int] = 0

    root: PyDataclassGenerator = field(init=False, default=None)

    parsed_types: TypeContainer = field(init=False,
                                        default_factory=TypeContainer)

    # Model is our model dataclass object, which may or may not be present
    # in the list. If there are multiple models (i.e. dicts), their keys
    # and the associated type defs should be merged into one model.
    model: PyDataclassGenerator = field(init=False, default=None)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: Optional[str]):
        """Title case and singularize the name."""
        if name:
            name = English.humanize(name)
            name = English.singularize(name).replace(' ', '')

        self._name = name

    def __post_init__(self, is_root: bool, nested_lvl: int):

        if not self.name:
            # Increment the suffix if needed
            if nested_lvl:
                self.name = f'{self.default_name}{nested_lvl}'
            else:
                self.name = self.default_name

        # Temp data dictionary object
        data_list = []

        for elem in self.data:

            typ = json_to_python_type(elem)

            if typ is PyDataType.DICT:

                typ = PyDataclassGenerator(elem, self.name,
                                           nested_lvl=nested_lvl,
                                           is_root=is_root)

                if self.model:
                    self.model |= typ
                    continue

                self.model = typ

            else:
                # Nested lists.
                if typ is PyDataType.LIST:
                    nested_lvl += 1
                    typ = PyListGenerator(elem, nested_lvl=nested_lvl)

                data_list.append(typ)

            self.parsed_types.append(typ)

        if is_root:

            # We want to start off by adding the nested `dataclass` field
            # first, so it shows up at the top of the container `dataclass`.
            data_dict = {self.name: self.model} if self.model else {}

            data_dict.update({
                f'field_{i + 1}': elem
                for i, elem in enumerate(data_list)
            })

            self.root = PyDataclassGenerator.load_parsed(
                data_dict,
                nested_lvl=nested_lvl
            )
            self.root.name = self.container_name

    def __or__(self, other):
        """Merge two lists together."""
        if not isinstance(other, PyListGenerator):
            raise TypeError(
                f'{self.__class__.__name__}: Incorrect type for `__or__`. '
                f'actual_type: {type(other)}, object={other}')

        # To merge lists with equal number of elements, that's easy enough:
        #   [{"key": "v1"}] | [{"key2": 2}] = [{"key": "v1", "key2": 2}]
        #
        # But... what happens when it's something like this?
        #   [1, {"key": "v1"}] | [{"key2": "2}, "testing", 1, 2, 3]
        #
        # Solution is to merge the model in the other list class with our
        # model -- note that both ours and the other instance end up with only
        # one model after `__post_init__` runs. However, easiest way is to
        # iterate over the nested types in the other list and check for the
        # model explicitly. For the rest of the types in the other list
        # (including nested lists), we just add them to our current list.
        for t in other.parsed_types:
            if isinstance(t, PyDataclassGenerator):
                if self.model:
                    self.model |= t
                    continue
                self.model = t
            self.parsed_types.append(t)

        return self

    def get_lines(self) -> List[str]:

        lines = []

        if self.root:
            lines.append(repr(self.root))

        else:
            if self.model:
                lines.append(repr(self.model))

            for t in self.parsed_types:
                if isinstance(t, PyListGenerator):
                    code = repr(t)
                    if code:
                        # Only if our list already has a dataclass, append
                        # a newline. This should add the proper number of
                        # spaces, in a case like below.
                        #   [{"another_Key":  "value"}, [{"key":  "value"}]]
                        if self.model:
                            lines.append('\n')
                        lines.append(code)

        return lines

    def __str__(self):
        ...

    def _default_str(self):

        if len(self.parsed_types) == 0:
            # We could also wrap it with 'Optional' here, since we see it's
            # an empty list, but it's probably better to not not do that, as
            # 'Optional' generally means the value can be an explicit "null".
            #
            # return ModuleImporter.wrap_string_with_import('list', Optional)
            return ModuleImporter.wrap_string_with_import('', List)

        return ModuleImporter.wrap_string_with_import(
            str(self.parsed_types), List)

    def _experimental_features_str(self):

        if len(self.parsed_types) == 0:
            return 'list'

        return ModuleImporter.wrap_string_with_import(
            str(self.parsed_types), list)

    def __repr__(self):
        """
        Returns the Python `dataclasses` representation of the object.
        """
        return '\n'.join(self.get_lines())


if __name__ == '__main__':
    loader = PyCodeGenerator('../../tests/testdata/test1.json')
    print(loader.py_code)


# Copyright (c) 2006 Bermi Ferrer Martinez
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software to deal in this software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of this software, and to permit
# persons to whom this software is furnished to do so, subject to the following
# condition:
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THIS SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THIS SOFTWARE.



"""
Entry point for the Wizard CLI tool.
"""
import argparse
import os
import platform
import sys
import textwrap
from gettext import gettext as _
from json import JSONDecodeError
from pathlib import Path
from typing import Optional, TextIO

from ..__version__ import __version__
from .schema import PyCodeGenerator

# Define the top-level parser
parser: argparse.ArgumentParser


def main(args=None):
    """
    A companion CLI tool for the Dataclass Wizard, which simplifies
    interaction with the Python `dataclasses` module.
    """

    setup_parser()

    args = parser.parse_args(args)

    try:
        args.func(args)

    except AttributeError:
        # A sub-command is not provided.
        parser.print_help()
        parser.exit(0)


def setup_parser():
    """Sets up the Wizard CLI parser."""
    global parser
    desc = main.__doc__
    py_version = sys.version.split(" ", 1)[0]

    # create the top-level parser
    parser = argparse.ArgumentParser(description=desc)

    # define global flags for the CLI tool
    parser.add_argument('-V', '--version', action='version',
                        version=f'%(prog)s-cli/{__version__} '
                                f'Python/{py_version} '
                                f'{platform.system()}/{platform.release()}',
                        help='Display the version of this tool.')
    # Commenting these out for now, as they are all currently a "no-op".
    # parser.add_argument('-v', '--verbose', action='store_true',
    #                     help='Enable verbose output')
    # parser.add_argument('-q', '--quiet', action='store_true')

    # Add the sub-commands here.

    subparsers = parser.add_subparsers(help='Supported sub-commands')

    # create the parser for the "gs" command
    gs_parser = subparsers.add_parser(
        'gen-schema', aliases=['gs'],
        help='Generates a Python dataclass schema, given a JSON input.')

    gs_parser.add_argument('in_file', metavar='in-file',
                           nargs='?',
                           type=FileTypeWithExt('r', ext='.json'),
                           help="Path to JSON file. The default assumes the "
                                "input is piped from stdin or '-'",
                           default=sys.stdin)

    gs_parser.add_argument('out_file', metavar='out-file',
                           nargs='?',
                           type=FileTypeWithExt('w', ext='.py'),
                           help="Path to new Python file. The default is to "
                                "print the output to stdout or '-'",
                           default=sys.stdout)

    gs_parser.add_argument("-n", "--no-json-file", action="store_true",
                           help='Do not create a separate JSON file. Note '
                                'this only applies when the JSON input is '
                                'piped in to stdin.')

    gs_parser.add_argument("-f", "--force-strings", action="store_true",
                           help='Force-resolve strings to inferred Python types. '
                                'For example, a string appearing as "TRUE" will '
                                'resolve to a `bool` type, instead of the '
                                'default `Union[bool, str]`.')

    gs_parser.add_argument("-x", "--experimental", action="store_true",
                           help='Enable experimental features via a __future__ '
                                'import, which allows PEP-585 and PEP-604 '
                                'style annotations in Python 3.7+')

    gs_parser.set_defaults(func=gen_py_schema)


class FileTypeWithExt(argparse.FileType):
    """
    Extends :class:`argparse.FileType` to add a default file extension if the
    provided file name is missing one.
    """

    def __init__(self, mode='r', ext=None,
                 bufsize=-1, encoding=None, errors='ignore'):

        super().__init__(mode, bufsize, encoding, errors)
        self._ext = ext

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == '-':
            if 'r' in self._mode:
                return sys.stdin
            elif 'w' in self._mode:  # pragma: no branch
                return sys.stdout
            else:   # pragma: no cover
                msg = _('argument "-" with mode %r') % self._mode
                raise ValueError(msg)

        # all other arguments are used as file names
        ext = os.path.splitext(string)[-1].lower()
        # Add the file extension, if needed
        if not ext and self._ext:
            string += self._ext
        try:
            return open(string, self._mode, self._bufsize, self._encoding,
                        self._errors)
        except OSError as e:
            message = _("can't open '%s': %s")
            raise argparse.ArgumentTypeError(message % (string, e))


def get_div(out_file: TextIO, char='_', line_width=50):
    """
    Returns a formatted line divider to print.
    """
    if out_file.isatty():
        try:
            w = os.get_terminal_size(out_file.fileno()).columns - 2
            if w > 0:
                line_width = w
        except (ValueError, OSError):
            # Perhaps not a real terminal after all
            pass

    return char * line_width


def gen_py_schema(args):
    """
    Entry point for the `wiz gen-schema (gs)` command.
    """

    in_file: TextIO = args.in_file
    out_file: TextIO = args.out_file
    no_json_file: bool = args.no_json_file
    force_strings: bool = args.force_strings
    experimental: bool = args.experimental

    # Currently these arguments are unused
    # verbose, quiet = args.verbose, args.quiet

    # Check if input is piped from stdin.
    is_stdin: bool = in_file.name == '<stdin>'

    # Check if output should be displayed to the terminal.
    is_stdout: bool = out_file.name == '<stdout>'

    # Read in contents of the JSON string, from stdin or a local file.
    json_string: str = in_file.read()

    try:
        code_gen = PyCodeGenerator(file_contents=json_string,
                                   force_strings=force_strings,
                                   experimental=experimental)

    except JSONDecodeError as e:
        msg = str(e).lower()

        if is_stdin and ('double quotes' in msg or 'extra data' in msg):
            # We can provide a more helpful error message in this case.
            msg = """\
            Confirm that double quotes are properly applied. For example, the following syntax is invalid:
                echo "{"key": "value"}" | wiz gs

            Instead, wrap the string with single quotes as shown below:
                echo \'{"key": "value"}\' | wiz gs
            """

            _exit_with_error(out_file, msg=msg)

        _exit_with_error(out_file, e)

    except Exception as e:
        _exit_with_error(out_file, e)

    else:
        print('Successfully generated the Python code for the JSON schema.')
        print(get_div(out_file))
        print()

        if not is_stdout:
            out_path = Path(out_file.name)
            # Only create the JSON file if we are piped the input, and the
            # `--no-json-file / -n` option is not passed in.
            add_json_file: bool = is_stdin and not no_json_file

            print(f'Wrote out the Python Code to:  {out_path.absolute()}')

            if add_json_file:
                json_loc = out_path.with_suffix('.json')
                json_loc.write_text(json_string)
                print(f'Saved the JSON Input to:       {json_loc.absolute()}')

        out_file.write(code_gen.py_code)


def _exit_with_error(out_file: TextIO,
                     e: Optional[Exception] = None,
                     msg: Optional[str] = None,
                     line_width=70,
                     indent='  '):
    """
    Prints the error message from an error `e` or an error message `msg`
    and exits the program.
    """

    msg_header = ('An error{err_cls}was encountered while parsing the JSON '
                  'input:')

    if not msg:
        msg = str(e)

    error_lines = [
        msg_header.format(err_cls=f' ({type(e).__name__}) ' if e else ' '),
        get_div(out_file)
    ]

    error_lines.extend(
        textwrap.wrap(
            textwrap.dedent(msg),
            width=line_width,
            initial_indent=indent,
            subsequent_indent=indent,
            drop_whitespace=False,
            replace_whitespace=False,
        )
    )

    sys.exit('\n'.join(error_lines))


if __name__ == "__main__":
    sys.exit(main())



import types
from collections import defaultdict, deque
from typing import TYPE_CHECKING, Any, TypedDict

from ._log import LOG
from ._type_def import META, DefFactory, NoneType, PyNotRequired
from ._type_utils import is_builtin
from .utils._function_builder import FunctionBuilder
from .utils._typing_compat import get_origin_v2

if TYPE_CHECKING:
    from .patterns import PatternBase


_BUILTIN_COLLECTION_TYPES = frozenset({
    list,
    set,
    dict,
    tuple,
    frozenset,
})

# FIXME: Python 3.9 doesn't have `types.EllipsisType` or `types.NotImplementedType`
EllipsisType = getattr(types, 'EllipsisType', type(Ellipsis))
NotImplementedType = getattr(types, 'NotImplementedType', type(NotImplemented))

LEAF_TYPES_NO_BYTES = frozenset({
    # Common JSON Serializable types
    NoneType,
    bool,
    int,
    float,
    str,
    # Other common types
    complex,
    # exclude bytes, since the serialization process is slightly different
    # Other types that are also unaffected by deepcopy
    EllipsisType,
    NotImplementedType,
    types.CodeType,
    types.BuiltinFunctionType,
    types.FunctionType,
    type,
    range,
    property,
})

# Atomic immutable types which don't require any recursive handling and for which deepcopy
# returns the same object. We can provide a fast-path for these types in asdict and astuple.
#
# Credits: `_ATOMIC_TYPES` from `dataclasses.py`
LEAF_TYPES = LEAF_TYPES_NO_BYTES | {bytes}

SEQUENCE_ORIGINS = frozenset({
    list,
    tuple,
    set,
    frozenset,
    deque
})

MAPPING_ORIGINS = frozenset({
    dict,
    defaultdict
})


class TypeInfo:

    __slots__ = (
        # type origin (ex. `List[str]` -> `List`)
        'origin',
        # type arguments (ex. `Dict[str, int]` -> `(str, int)`)
        'args',
        # name of type origin (ex. `List[str]` -> 'list')
        'name',
        # index of iteration, *only* unique within the scope of a field assignment!
        'i',
        # index of field within the dataclass, *guaranteed* to be unique.
        'field_i',
        # prefix of value in assignment (prepended to `i`),
        # defaults to 'v' if not specified.
        'prefix',
        # index of assignment (ex. `2 -> v1[2]`, *or* a string `"key" -> v4["key"]`)
        'index',
        # explicit value name (overrides prefix + index)
        'val_name',
        # optional attribute, that indicates if we should wrap the
        # assignment with `name` -- ex. `(1, 2)` -> `deque((1, 2))`
        '_wrapped',
        # optional attribute, that indicates if we are currently in Optional,
        # e.g. `typing.Optional[...]` *or* `typing.Union[T, ...*T2, None]`
        '_in_opt',
    )

    def __init__(self, origin,
                 args=None,
                 name=None,
                 i=1,
                 field_i=1,
                 prefix='v',
                 val_name=None,
                 index=None):

        self.name = name
        self.origin = origin
        self.args = args
        self.i = i
        self.field_i = field_i
        self.prefix = prefix
        self.val_name = val_name
        self.index = index

    def replace(self, **changes):
        # Validate that `instance` is an instance of the class
        # if not isinstance(instance, TypeInfo):
        #     raise TypeError(f"Expected an instance of {TypeInfo.__name__}, got {type(instance).__name__}")

        # Extract current values from __slots__
        current_values = {slot: getattr(self, slot)
                          for slot in TypeInfo.__slots__
                          if not slot.startswith('_')}


        if ((new_idx := changes.get('index')) is not None
            and (curr_idx := current_values['index']) is not None):
            if isinstance(curr_idx, (int, str)):
                changes['index'] = (curr_idx, new_idx)
            else:
                changes['index'] = curr_idx + (new_idx, )

        # Apply the changes
        current_values.update(changes)

        # Create and return a new instance with updated attributes
        # noinspection PyArgumentList
        return TypeInfo(**current_values)

    @property
    def in_optional(self):
        return getattr(self, '_in_opt', False)

    # noinspection PyUnresolvedReferences
    @in_optional.setter
    def in_optional(self, value):
        # noinspection PyAttributeOutsideInit
        self._in_opt = value

    @staticmethod
    def ensure_in_locals(extras, *tps, **name_to_tp):
        names = [ensure_type_ref(extras, tp) for tp in tps]

        for name, tp in name_to_tp.items():
            extras['locals'].setdefault(name, tp)

        return names

    def type_name(self, extras, bound=None):
        """Return type name as string (useful for `Union` type checks)"""
        if self.name is None:
            self.name = get_origin_v2(self.origin).__name__

        return self._wrap_inner(
            extras, force=True, bound=bound)

    def v(self):
        val_name = self.val_name
        if val_name is None:
            val_name = f'{self.prefix}{self.i}'
        idx = self.index
        if idx is None:
            return val_name
        else:
            if isinstance(idx, (int, str)):
                return f'{val_name}[{idx}]'
            return f"{val_name}{''.join(f'[{i}]' for i in idx)}"

    def v_for_def(self):
        """
        Returns a safe value for function `def` statements (e.g., no
        dot (.) or indices [])
        """
        return f'{self.prefix}{self.i}'

    def v_and_next(self):
        next_i = self.i + 1
        return self.v(), f'{self.prefix}{next_i}', next_i

    def v_and_next_k_v(self):
        next_i = self.i + 1
        return self.v(), f'k{next_i}', f'v{next_i}', next_i

    def wrap_dd(self, default_factory: DefFactory, result: str, extras):
        tn = self._wrap_inner(extras, is_builtin=True, bound=defaultdict)
        tn_df = self._wrap_inner(extras, default_factory)
        result = f'{tn}({tn_df}, {result})'
        self._wrapped = result
        return self

    def multi_wrap(self, extras, prefix='', *result, force=False):
        tn = self._wrap_inner(extras, prefix=prefix, force=force)
        if tn is not None:
            result = [f'{tn}({r})' for r in result]

        return result

    def wrap(self, result: str, extras, force=False, prefix='', bound=None):
        tn = self._wrap_inner(extras, prefix=prefix, force=force, bound=bound)
        if tn is not None:
            result = f'{tn}({result})'

        self._wrapped = result
        return self

    def wrap_builtin(self, bound, result, extras):
        tn = self._wrap_inner(extras, is_builtin=True, bound=bound)
        result = f'{tn}({result})'

        self._wrapped = result
        return self

    def _wrap_inner(self, extras,
                    tp=None,
                    prefix='',
                    is_builtin=False,
                    force=False,
                    bound=None) -> 'str | None':

        if tp is None:
            tp = self.origin
            name = self.name
            return_name = force
        else:
            name = 'None' if tp is NoneType else tp.__name__
            return_name = True

        # If the type is the bound itself, treat it as "builtin" in naming
        # (i.e., don't generate unique alias)
        #
        # This ensures we don't create a "unique" name
        # if it's a non-subclass, e.g. ensures we end
        # up with `date` instead of `date_123`.
        if bound is not None:
            is_builtin = tp is bound

        if tp not in _BUILTIN_COLLECTION_TYPES:
            return ensure_type_ref(
                extras,
                tp,
                name=name,
                prefix=prefix,
                is_builtin=is_builtin,
            )

        return name if return_name else None

    def __str__(self):
        return getattr(self, '_wrapped', '')

    def __repr__(self):  # pragma: no cover
        items = ', '.join([f'{v}={getattr(self, v)!r}'
                           for v in self.__slots__
                           if not v.startswith('_')])

        return f'{self.__class__.__name__}({items})'


class Extras(TypedDict):
    """
    "Extra" config that can be used in the load / dump process.
    """
    config: 'META'
    cls: type
    cls_name: str
    fn_gen: FunctionBuilder
    locals: dict[str, Any]
    pattern: PyNotRequired['PatternBase']
    recursion_guard: dict[type, str]


def ensure_type_ref(extras, tp, *, name=None, prefix='', is_builtin=False) -> str:
    """
    Return a safe symbol name for `tp` to use in generated code.

    Adds entries to `extras['locals']` only when required (non-builtins,
    non-collection literals, and cases where a stable local alias is needed).
    """
    if tp is NoneType:
        return 'None'

    if name is None:
        name = tp.__name__

    # Common built-in collections: always use the literal names directly.
    if tp in _BUILTIN_COLLECTION_TYPES:
        return name

    mod = tp.__module__

    # Builtins: can be referenced directly without injecting into locals.
    # Includes str/int/float/bool/bytes and also built-in collection types.
    if mod == 'builtins':
        return name

    if is_builtin or mod == 'collections':
        LOG.debug('Ensuring %s=%s', name, name)
        extras['locals'].setdefault(name, tp)
        return name

    _locals = extras['locals']

    # If the type name is safe and not used yet, inject it.
    # You may want stricter collision checks here.
    if name not in _locals:
        _locals[name] = tp
        return name

    # Collision: create a unique alias.
    # TODO might need to handle `var_name`
    alias = f'{prefix}{name}'
    LOG.debug('Adding %s=%s', alias, name)
    _locals.setdefault(alias, tp)

    return alias


def finalize_skip_if(skip_if, operand_1, conditional):
    """
    Finalizes the skip condition by generating the appropriate string based on the condition.

    Args:
        skip_if (Condition): The condition to evaluate, containing truthiness and operation info.
        operand_1 (str): The primary operand for the condition (e.g., a variable or value).
        conditional (str): The conditional operator to use (e.g., '==', '!=').

    Returns:
        str: The resulting skip condition as a string.

    Example:
        >>> from dataclass_wizard.conditions import Condition
        >>> cond = Condition(t_or_f=True, op='+', val=None)
        >>> finalize_skip_if(cond, 'my_var', '==')
        'my_var'
    """
    if skip_if.t_or_f:
        return operand_1 if skip_if.op == '+' else f'not {operand_1}'

    return f'{operand_1} {conditional}'


def get_skip_if_condition(skip_if, _locals, operand_2=None, condition_i=None, condition_var='_skip_if_'):
    """
    Retrieves the skip condition based on the provided `Condition` object.

    Args:
        skip_if (Condition): The condition to evaluate.
        _locals (dict[str, Any]): A dictionary of local variables for condition evaluation.
        operand_2 (str): The secondary operand (e.g., a variable or value).
        condition_i (Condition): The condition var index.
        condition_var (str): The variable name to evaluate.

    Returns:
        Any: The result of the evaluated condition or a string representation for custom values.

    Example:
        >>> from dataclass_wizard.conditions import Condition
        >>> cond = Condition(t_or_f=False, op='==', val=10)
        >>> locals_dict = {}
        >>> get_skip_if_condition(cond, locals_dict, 'other_var')
        '== other_var'
    """
    if skip_if is None:
        return False

    if skip_if.t_or_f:  # Truthy or falsy condition, no operand
        return True

    if is_builtin(skip_if.val):
        return str(skip_if)

    # Update locals (as `val` is not a builtin)
    if operand_2 is None:
        operand_2 = f'{condition_var}{condition_i}'

    _locals[operand_2] = skip_if.val
    return f'{skip_if.op} {operand_2}'



"""
Ideally should be in the `bases` module, however we'll run into a Circular
Import scenario if we move it there, since the `loaders` and `dumpers` modules
both import directly from `bases`.

"""
from __future__ import annotations

import logging
from collections.abc import Mapping

from ._bases import AbstractEnvMeta, AbstractMeta
from ._class_helper import (
    DATACLASS_FIELD_TO_ALIAS_FOR_DUMP,
    DATACLASS_FIELD_TO_ALIAS_FOR_LOAD,
    DATACLASS_FIELD_TO_ENV_FOR_LOAD,
    META_INITIALIZER,
)
from ._dumpers import DumpMixin, get_dumper
from ._loaders import LoadMixin, get_loader
from ._log import LOG
from ._meta_cache import META_BY_DATACLASS, get_meta, set_base_meta_cls
from ._type_conv import as_enum
from ._type_def import E
from ._type_utils import (
    create_new_class,
    get_class_name,
    get_outer_class_name,
    per_cls,
)
from .enums import DateTimeTo, EnvKeyStrategy, EnvPrecedence, KeyAction, KeyCase
from .errors import ParseError

ALLOWED_MODES = ('runtime', 'codegen')

# global flag to determine if debug mode was ever enabled
_debug_was_enabled = False


def register_type(cls, tp, *, load=None, dump=None, mode=None) -> None:
    meta = get_meta(cls)
    if meta is AbstractMeta:
        from ._meta_cache import create_meta
        meta = create_meta(cls)

    if load is None:
        load = tp
    if dump is None:
        dump = str

    if (load_hook := meta.type_to_load_hook) is None:
        meta.type_to_load_hook = load_hook = {}
    if (dump_hook := meta.type_to_dump_hook) is None:
        meta.type_to_dump_hook = dump_hook = {}

    load_hook[tp] = (mode if mode else _infer_mode(load), load)
    dump_hook[tp] = (mode if mode else _infer_mode(dump), dump)


# use `debug` for log level if it's a str or int.
def _enable_debug_mode_if_needed(possible_lvl):
    global _debug_was_enabled
    if not _debug_was_enabled:
        _debug_was_enabled = True
        # use `debug` for log level if it's a str or int.
        default_lvl = logging.DEBUG
        # minimum logging level for logs by this library.
        min_level = default_lvl if isinstance(
            possible_lvl, bool) else possible_lvl
        # set the logging level of this library's logger.
        LOG.setLevel(min_level)
        LOG.info('DEBUG Mode is enabled')


def _as_enum_safe(cls: type, name: str, base_type: type[E]) -> E | None:
    """
    Attempt to return the value for class attribute :attr:`attr_name` as
    a :type:`base_type`.

    :raises ParseError: If we are unable to convert the value of the class
      attribute to an Enum of type `base_type`.
    """
    try:
        return as_enum(getattr(cls, name), base_type)

    except ParseError as e:
        # We run into a parsing error while loading the enum; Add
        # additional info on the Exception object before re-raising it
        e.class_name = get_class_name(cls)
        e.field_name = name
        raise


def _infer_mode(hook) -> str:
    code = getattr(hook, '__code__', None)

    if code is None:
        return 'runtime'  # types/builtins

    co_flags = code.co_flags
    if co_flags & 0x04 or co_flags & 0x08:
        raise TypeError('hooks must not use *args/**kwargs')

    argc = code.co_argcount
    if argc == 1:
        return 'runtime'
    if argc == 2:
        return 'codegen'

    raise TypeError('hook must accept 1 arg (runtime) '
                    'or 2 args (TypeInfo, Extras)')


def _normalize_hooks(hooks: Mapping | None) -> None:
    if not hooks:
        return

    for tp, hook in hooks.items():
        if isinstance(hook, tuple):
            if len(hook) != 2:
                raise ValueError(
                    'hook tuple must be (mode, hook), '
                    f'got {hook!r}') from None

            mode, fn = hook
            if mode not in ALLOWED_MODES:
                raise ValueError(
                    f"mode must be 'runtime' or 'codegen' (got {mode!r})"
                ) from None

        else:
            mode = _infer_mode(hook)
            # noinspection PyUnresolvedReferences
            hooks[tp] = mode, hook


class BaseJSONWizardMeta(AbstractMeta):
    """
    Superclass definition for the `JSONWizard.Meta` inner class.

    See the implementation of the :class:`AbstractMeta` class for the
    available config that can be set, as well as for descriptions on any
    implemented methods.
    """

    __slots__ = ()

    @classmethod
    def _init_subclass(cls):
        """
        Hook that should ideally be run whenever the `Meta` class is
        sub-classed.

        """
        outer_cls_name = get_outer_class_name(cls, raise_=False)

        # We can retrieve the outer class name using `__qualname__`, but it's
        # not easy to find the class definition itself. The simplest way seems
        # to be to create a new callable (essentially a class method for the
        # outer class) which will later be called by the base enclosing class.
        #
        # Note that this relies on the observation that the
        # `__init_subclass__` method of any inner classes are run before the
        # one for the outer class.
        if outer_cls_name is not None:
            META_INITIALIZER[outer_cls_name] = cls.bind_to
        else:
            # The `Meta` class is defined as an outer class. Emit a warning
            # here, just so we can ensure awareness of this special case.
            LOG.warning('The %r class is not declared as an Inner Class, so '
                        'these are global settings that will apply to all '
                        'JSONSerializable sub-classes.', get_class_name(cls))

            # Copy over global defaults to the :class:`AbstractMeta`
            for attr in AbstractMeta.fields_to_merge:
                setattr(AbstractMeta, attr, getattr(cls, attr, None))
            if cls.field_to_alias:
                AbstractMeta.field_to_alias = cls.field_to_alias
            if cls.field_to_alias_dump:
                AbstractMeta.field_to_alias_dump = cls.field_to_alias_dump
            if cls.field_to_alias_load:
                AbstractMeta.field_to_alias_load = cls.field_to_alias_load

            # Create a new class of `Type[W]`, and then pass `create=False` so
            # that we don't create new loader / dumper for the class.
            new_cls = create_new_class(cls, ())
            cls.bind_to(new_cls, create=False)

    @classmethod
    def bind_to(cls, dataclass: type, create=True, is_default=True,
                base_loader=LoadMixin,
                base_dumper=DumpMixin):
        cls_loader = get_loader(dataclass, create=create,
                                base_cls=base_loader)
        cls_dumper = get_dumper(dataclass, create=create,
                                base_cls=base_dumper)

        if cls.debug:
            _enable_debug_mode_if_needed(cls.debug)

        if cls.dump_date_time_as is not None:
            cls.dump_date_time_as = _as_enum_safe(
                cls, 'dump_date_time_as', DateTimeTo)

        if (key_case := cls.case) is not None:
            cls.load_case = cls.dump_case = key_case
            cls.case = None

        if cls.load_case is not None:
            cls_loader.transform_json_field = _as_enum_safe(
                cls, 'load_case', KeyCase)

        if cls.dump_case is not None:
            cls_dumper.transform_dataclass_field = _as_enum_safe(
                cls, 'dump_case', KeyCase)

        if (field_to_alias := cls.field_to_alias) is not None:
            cls.field_to_alias_dump = {
                k: v if isinstance(v, str) else v[0]
                for k, v in field_to_alias.items()
            }
            cls.field_to_alias_load = field_to_alias

        if (field_to_alias := cls.field_to_alias_dump) is not None:
            per_cls(DATACLASS_FIELD_TO_ALIAS_FOR_DUMP, dataclass).update(
                field_to_alias)

        if (field_to_alias := cls.field_to_alias_load) is not None:
            per_cls(DATACLASS_FIELD_TO_ALIAS_FOR_LOAD, dataclass).update({
                k: (v, ) if isinstance(v, str) else v
                for k, v in field_to_alias.items()
            })

        if cls.on_unknown_key is not None:
            cls.on_unknown_key = _as_enum_safe(
                cls, 'on_unknown_key', KeyAction)

        _normalize_hooks(cls.type_to_load_hook)
        _normalize_hooks(cls.type_to_dump_hook)

        # Finally, if needed, save the meta config for the outer class. This
        # will allow us to access this config as part of the JSON load/dump
        # process if needed.
        if is_default:
            # Check if the dataclass already has a Meta config; if so, we
            # need to copy over special attributes so they don't get
            # overwritten.
            if dataclass in META_BY_DATACLASS:
                META_BY_DATACLASS[dataclass] &= cls
            else:
                META_BY_DATACLASS[dataclass] = cls


# IMPORTANT: do this after the class definition
set_base_meta_cls(BaseJSONWizardMeta)


class BaseEnvWizardMeta(AbstractEnvMeta):
    """
    Superclass definition for the `EnvWizard.Meta` inner class.

    See the implementation of the :class:`AbstractEnvMeta` class for the
    available config that can be set, as well as for descriptions on any
    implemented methods.
    """

    __slots__ = ()

    @classmethod
    def _init_subclass(cls):
        """
        Hook that should ideally be run whenever the `Meta` class is
        sub-classed.

        """
        outer_cls_name = get_outer_class_name(cls, raise_=False)

        if outer_cls_name is not None:
            META_INITIALIZER[outer_cls_name] = cls.bind_to
        else:
            # The `Meta` class is defined as an outer class. Emit a warning
            # here, just so we can ensure awareness of this special case.
            LOG.warning('The %r class is not declared as an Inner Class, so '
                        'these are global settings that will apply to all '
                        'EnvWizard sub-classes.', get_class_name(cls))

            # Copy over global defaults to the :class:`AbstractMeta`
            for attr in AbstractEnvMeta.fields_to_merge:
                setattr(AbstractEnvMeta, attr, getattr(cls, attr, None))
            if cls.field_to_alias_dump:
                AbstractEnvMeta.field_to_alias_dump = cls.field_to_alias_dump
            if cls.field_to_env_load:
                AbstractEnvMeta.field_to_env_load = cls.field_to_env_load

            # Create a new class of `Type[W]`, and then pass `create=False` so
            # that we don't create new loader / dumper for the class.
            new_cls = create_new_class(cls, ())
            cls.bind_to(new_cls, create=False)

    @classmethod
    def bind_to(cls, env_class: type, create=True, is_default=True):
        cls_dumper = get_dumper(
            env_class,
            create=create)

        if cls.debug:
            _enable_debug_mode_if_needed(cls.debug)

        if cls.load_case is not None:
            cls.load_case = _as_enum_safe(
                cls, 'load_case', EnvKeyStrategy)
        if cls.env_precedence is not None:
            cls.env_precedence = _as_enum_safe(
                cls, 'env_precedence', EnvPrecedence)

        # TODO
        cls_dumper.transform_dataclass_field = _as_enum_safe(
            cls, 'dump_case', KeyCase)

        if (field_to_alias := cls.field_to_alias_dump) is not None:
            per_cls(DATACLASS_FIELD_TO_ALIAS_FOR_DUMP, env_class).update(
                field_to_alias)

        if (field_to_env := cls.field_to_env_load) is not None:
            per_cls(DATACLASS_FIELD_TO_ENV_FOR_LOAD, env_class).update({
                k: (v, ) if isinstance(v, str) else v
                for k, v in field_to_env.items()
            })

        # set this attribute in case of nested dataclasses (which
        # uses codegen in `loaders.py`)
        cls.on_unknown_key = None

        # if cls.on_unknown_key is not None:
        #     cls.on_unknown_key = _as_enum_safe(
        #       cls, 'on_unknown_key', KeyAction)

        _normalize_hooks(cls.type_to_load_hook)
        _normalize_hooks(cls.type_to_dump_hook)

        # Finally, if needed, save the meta config for the outer class. This
        # will allow us to access this config as part of the JSON load/dump
        # process if needed.
        if is_default:
            # Check if the dataclass already has a Meta config; if so, we
            # need to copy over special attributes so they don't get
            # overwritten.
            if env_class in META_BY_DATACLASS:
                META_BY_DATACLASS[env_class] &= cls
            else:
                META_BY_DATACLASS[env_class] = cls


# noinspection PyPep8Naming, PyUnresolvedReferences
def LoadMeta(**kwargs):
    """
    Helper function to setup the ``Meta`` Config for the JSON load
    (de-serialization) process, which is intended for use alongside the
    ``fromdict`` helper function.

    For descriptions on what each of these params does, refer to the `Docs`_
    below, or check out the :class:`AbstractMeta` definition (I want to avoid
    duplicating the descriptions for params here).

    Examples::

        >>> from dataclass_wizard import LoadMeta, fromdict
        >>> LoadMeta(key_transform='CAMEL').bind_to(MyClass)
        >>> fromdict(MyClass, {"myStr": "value"})

    .. _Docs: https://dcw.ritviknag.com/en/latest/common_use_cases/meta.html
    """
    base_dict = kwargs | {'__slots__': ()}

    if (v := base_dict.pop('key_transform', None)) is not None:
        base_dict['key_transform_with_load'] = v

    if (v := base_dict.pop('case', None)) is not None:
        base_dict['load_case'] = v

    if (v := base_dict.pop('field_to_alias', None)) is not None:
        base_dict['field_to_alias_load'] = v

    if (v := base_dict.pop('type_to_hook', None)) is not None:
        base_dict['type_to_load_hook'] = v

    # Create a new subclass of :class:`AbstractMeta`
    # noinspection PyTypeChecker
    return type('Meta', (BaseJSONWizardMeta, ), base_dict)


# noinspection PyPep8Naming, PyUnresolvedReferences
def DumpMeta(**kwargs):
    """
    Helper function to setup the ``Meta`` Config for the JSON dump
    (serialization) process, which is intended for use alongside the
    ``asdict`` helper function.

    For descriptions on what each of these params does, refer to the `Docs`_
    below, or check out the :class:`AbstractMeta` definition (I want to avoid
    duplicating the descriptions for params here).

    Examples::

        >>> from dataclass_wizard import DumpMeta, asdict
        >>> DumpMeta(key_transform='CAMEL').bind_to(MyClass)
        >>> asdict(MyClass, {"myStr": "value"})

    .. _Docs: https://dcw.ritviknag.com/en/latest/common_use_cases/meta.html
    """

    # Set meta attributes here.
    base_dict = kwargs | {'__slots__': ()}

    if (v := base_dict.pop('key_transform', None)) is not None:
        base_dict['key_transform_with_dump'] = v

    if (v := base_dict.pop('case', None)) is not None:
        base_dict['dump_case'] = v

    if (v := base_dict.pop('field_to_alias', None)) is not None:
        base_dict['field_to_alias_dump'] = v

    if (v := base_dict.pop('type_to_hook', None)) is not None:
        base_dict['type_to_dump_hook'] = v

    # Create a new subclass of :class:`AbstractMeta`
    # noinspection PyTypeChecker
    return type('Meta', (BaseJSONWizardMeta, ), base_dict)


# noinspection PyPep8Naming, PyUnresolvedReferences
def EnvMeta(**kwargs):
    """
    Helper function to setup the ``Meta`` Config for the EnvWizard.

    For descriptions on what each of these params does, refer to the `Docs`_
    below, or check out the :class:`AbstractEnvMeta` definition (I want
    to avoid duplicating the descriptions for params here).

    Examples::

        >>> EnvMeta(key_transform_with_dump='SNAKE').bind_to(MyClass)

    .. _Docs: https://dcw.ritviknag.com/en/latest/common_use_cases/meta.html
    """

    # Set meta attributes here.
    base_dict = kwargs | {'__slots__': ()}

    # Create a new subclass of :class:`AbstractMeta`
    # noinspection PyTypeChecker
    return type('EnvMeta', (BaseEnvWizardMeta, ), base_dict)



from typing import AnyStr, BinaryIO, TextIO

from .._serial_json import SerializerHookMixin
from .._type_def import Decoder, Encoder, FileDecoder, FileEncoder, FileType, T
from ..enums import KeyCase

class YAMLWizard(SerializerHookMixin):

    def __init_subclass__(cls, dump_case=KeyCase.KEBAB):
        ...

    @classmethod
    def from_yaml(cls: type[T],
                  string_or_stream: AnyStr | TextIO | BinaryIO, *,
                  decoder: Decoder | None = None,
                  **decoder_kwargs) -> T | list[T]:
        ...

    @classmethod
    def from_yaml_file(cls: type[T], file: FileType, *,
                       decoder: FileDecoder | None = None,
                       **decoder_kwargs) -> T | list[T]:
        ...

    def to_yaml(self: T, *,
                encoder: Encoder | None = None,
                **encoder_kwargs) -> str:
        ...

    def to_yaml_file(self: T, file: FileType, mode: str = 'w',
                     encoder: FileEncoder | None = None,
                     **encoder_kwargs) -> None:
        ...

    @classmethod
    def list_to_yaml(cls: type[T],
                     instances: list[T],
                     encoder: Encoder | None = None,
                     **encoder_kwargs) -> str:
        ...



"""
Pulling some functions removed in recent versions of Python into the module for continued compatibility.
All function names and bodies are left exactly as they were prior to being removed.
"""

from dataclasses import MISSING, dataclass, fields, is_dataclass
from types import FunctionType
from weakref import WeakKeyDictionary

from ..constants import PY310_OR_ABOVE

FIELDS = WeakKeyDictionary()
SEEN_DEFAULT = WeakKeyDictionary()


def set_qualname(cls, value):
    # Removed in Python 3.13
    # Original: `dataclasses._set_qualname`
    # Ensure that the functions returned from _create_fn uses the proper
    # __qualname__ (the class they belong to).
    if isinstance(value, FunctionType):
        value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
    return value


def set_new_attribute(cls, name, value, force=False):
    # Removed in Python 3.13
    # Original: `dataclasses._set_new_attribute`
    # Never overwrites an existing attribute.  Returns True if the
    # attribute already exists.
    if force or name not in cls.__dict__:
        set_qualname(cls, value)
        setattr(cls, name, value)
        return False
    return True


# noinspection PyShadowingBuiltins
def create_fn(name, args, body, *, globals=None, locals=None,
              return_type=MISSING):
    # Removed in Python 3.13
    # Original: `dataclasses._create_fn`
    # Note that we may mutate locals. Callers beware!
    # The only callers are internal to this module, so no
    # worries about external callers.
    if locals is None:
        locals = {}
    return_annotation = ''
    if return_type is not MISSING:
        locals['__dataclass_return_type__'] = return_type
        return_annotation = '->__dataclass_return_type__'
    args = ','.join(args)
    body = '\n'.join(f'  {b}' for b in body)

    # Compute the text of the entire function.
    txt = f' def {name}({args}){return_annotation}:\n{body}'

    # Free variables in exec are resolved in the global namespace.
    # The global namespace we have is user-provided, so we can't modify it for
    # our purposes. So we put the things we need into locals and introduce a
    # scope to allow the function we're creating to close over them.
    local_vars = ', '.join(locals.keys())
    txt = f"def __create_fn__({local_vars}):\n{txt}\n return {name}"
    ns = {}
    exec(txt, globals, ns)
    return ns['__create_fn__'](**locals)


def dataclass_needs_refresh(cls) -> bool:
    if not is_dataclass(cls):
        return True

    # dataclass fields currently registered
    # noinspection PyDataclass
    dc_fields = {f.name for f in fields(cls)}
    # annotated fields declared on the class (ignore ClassVar/InitVar nuance)
    ann = getattr(cls, '__annotations__', {}) or {}
    annotated = set(ann.keys())

    # If class declares annotated fields not present in dataclass fields,
    # the dataclass metadata is stale.
    return not annotated.issubset(dc_fields)


if PY310_OR_ABOVE:
    def apply_env_wizard_dataclass(cls, dc_kwargs):
        # noinspection PyArgumentList
        return dataclass(
            cls,
            init=False,
            kw_only=True,
            **dc_kwargs,
        )

    def dataclass_kw_only_init_field_names(cls):
        return {f.name for f in dataclass_init_fields(cls) if f.kw_only}

else:  # Python 3.9: no `kw_only`
    # noinspection PyArgumentList
    def apply_env_wizard_dataclass(cls, dc_kwargs):
        return dataclass(
            cls,
            init=False,
            **dc_kwargs,
        )

    def dataclass_kw_only_init_field_names(_):
        return set()


def dataclass_fields(cls):
    try:
        return FIELDS[cls]
    except KeyError:
        # noinspection PyTypeChecker,PyDataclass
        FIELDS[cls] = fs = fields(cls)
        return fs


def dataclass_init_fields(cls, as_list=False):
    init_fields = [f for f in dataclass_fields(cls) if f.init]
    return init_fields if as_list else tuple(init_fields)


def dataclass_field_names(cls):
    return tuple(f.name for f in dataclass_fields(cls))


def dataclass_init_field_names(cls):
    return tuple(f.name for f in dataclass_init_fields(cls))


def str_pprint_fn():
    from pprint import pformat
    return create_fn(
        '__str__',
        ('self',),
        [
            'try:',
            '    return pformat(self.to_dict(), width=70)',
            'except Exception:',
            '    return object.__repr__(self)',
        ],
        globals={'pformat': pformat},
    )



"""
JSON Helper Utilities - *only* internally used in ``errors.py``,
i.e. for rendering exceptions.

.. NOTE::
    This module should not be imported anywhere at the *top-level*
    of another library module!

"""
__all__ = [
    'safe_dumps',
]

from dataclasses import is_dataclass
from datetime import datetime, time, date
from enum import Enum
from json import dumps, JSONEncoder
from typing import Any
from uuid import UUID

from ..loader_selection import asdict


class SafeEncoder(JSONEncoder):
    """
    A Customized JSON Encoder, which copies core logic in the
    `dumpers` module to support serialization of more complex
    Python types, such as `datetime` and `Enum`.
    """

    def default(self, o: Any) -> Any:
        """Default function, copies the core (minimal) logic from `dumpers.py`."""

        if is_dataclass(o):
            return asdict(o)

        if isinstance(o, Enum):
            return o.value

        if isinstance(o, UUID):
            return o.hex

        if isinstance(o, (datetime, time)):
            return o.isoformat().replace('+00:00', 'Z', 1)

        if isinstance(o, date):
            return o.isoformat()

        # anything else (Decimal, timedelta, etc.)
        return str(o)


def safe_dumps(o, cls=SafeEncoder, **kwargs):
    try:
        return dumps(o, cls=cls, **kwargs)
    except TypeError:
        return o



from collections import namedtuple
from dataclasses import dataclass
from typing import NamedTuple

from dataclass_wizard import DataclassWizard, EnvWizard

from tests._typing import Required, NotRequired, ReadOnly, TypedDict


class TNReq(NamedTuple):
    a: int
    b: int


class TN(NamedTuple):
    a: int
    b: int = 2


CN = namedtuple("CN", "a b", defaults=(2,))


# 1) total=False + Required/NotRequired/ReadOnly, E2E
class TD_TF(TypedDict, total=False):
    a: Required[int]          # required even though total=False
    b: NotRequired[int]       # optional
    ro: ReadOnly[int]         # optional because total=False


class ContTF(DataclassWizard):
    td: TD_TF


class EnvContTF(EnvWizard):
    td: TD_TF


# 2) total=True + NotRequired + ReadOnly requiredness, E2E
class TD_TT(TypedDict):
    a: int                    # required
    b: NotRequired[int]       # optional
    ro: ReadOnly[int]         # required (unless NotRequired wraps it)


class ContTT(DataclassWizard):
    td: TD_TT


class EnvContTT(EnvWizard):
    td: TD_TT


# 3) all-required TypedDict (no optional keys) => codegen inline path, E2E
class TD_AllReq(TypedDict):
    x: int
    y: str


class ContAllReq(DataclassWizard):
    td: TD_AllReq


class EnvContAllReq(EnvWizard):
    td: TD_AllReq


@dataclass
class Sub2:
    my_float: float
